import numpy as np

from depthai_sdk.classes import Detections


def decode(nn_data):
    """
    Each palm detection is a tensor consisting of 19 numbers:
        - ymin, xmin, ymax, xmax
        - x,y-coordinates for the 7 key_points
        - confidence score
    :return:
    """
    if nn_data is None:
        return Detections(nn_data)

    shape = (128, 128)
    num_keypoints = 7
    min_score_thresh = 0.7
    anchors = np.load(__file__.replace("handler.py", "anchors_palm.npy"))

    # Run the neural network
    results = to_tensor_result(nn_data)

    raw_box_tensor = results.get("regressors").reshape(-1, 896, 18)  # regress
    raw_score_tensor = results.get("classificators").reshape(-1, 896, 1)  # classification

    detections = raw_to_detections(raw_box_tensor, raw_score_tensor, anchors, shape, num_keypoints)

    palm_coords = [
        obj[:4]
        for det in detections
        for obj in det
        if obj[-1] > min_score_thresh
    ]
    palm_confs = [
        obj[-1] for det in detections for obj in det if obj[-1] > min_score_thresh
    ]

    if len(palm_coords) == 0:
        return Detections(nn_data)

    nms = non_max_suppression(
        boxes=np.concatenate(palm_coords).reshape(-1, 4),
        probs=palm_confs,
        overlap_threshold=0.1
    )

    if nms is None:
        return Detections(nn_data)

    detections = Detections(nn_data)
    for i, box in enumerate(nms):
        detections.add(0, palm_confs[i], tuple(box))

    return detections


def sigmoid(x):
    return (1.0 + np.tanh(0.5 * x)) * 0.5


def decode_boxes(raw_boxes, anchors, shape, num_keypoints):
    """
    Converts the predictions into actual coordinates using the anchor boxes.
    Processes the entire batch at once.
    """
    boxes = np.zeros_like(raw_boxes)
    x_scale, y_scale = shape

    x_center = raw_boxes[..., 0] / x_scale * anchors[:, 2] + anchors[:, 0]
    y_center = raw_boxes[..., 1] / y_scale * anchors[:, 3] + anchors[:, 1]

    w = raw_boxes[..., 2] / x_scale * anchors[:, 2]
    h = raw_boxes[..., 3] / y_scale * anchors[:, 3]

    boxes[..., 1] = y_center - h / 2.0  # xmin
    boxes[..., 0] = x_center - w / 2.0  # ymin
    boxes[..., 3] = y_center + h / 2.0  # xmax
    boxes[..., 2] = x_center + w / 2.0  # ymax

    for k in range(num_keypoints):
        offset = 4 + k * 2
        keypoint_x = raw_boxes[..., offset] / x_scale * anchors[:, 2] + anchors[:, 0]
        keypoint_y = (
                raw_boxes[..., offset + 1] / y_scale * anchors[:, 3] + anchors[:, 1]
        )
        boxes[..., offset] = keypoint_x
        boxes[..., offset + 1] = keypoint_y

    return boxes


def raw_to_detections(raw_box_tensor, raw_score_tensor, anchors_, shape, num_keypoints):
    """

    This function converts these two "raw" tensors into proper detections.
    Returns a list of (num_detections, 17) tensors, one for each image in
    the batch.

    This is based on the source code from:
    mediapipe/calculators/tflite/tflite_tensors_to_detections_calculator.cc
    mediapipe/calculators/tflite/tflite_tensors_to_detections_calculator.proto
    """
    detection_boxes = decode_boxes(raw_box_tensor, anchors_, shape, num_keypoints)
    detection_scores = sigmoid(raw_score_tensor).squeeze(-1)
    output_detections = []
    for i in range(raw_box_tensor.shape[0]):
        boxes = detection_boxes[i]
        scores = np.expand_dims(detection_scores[i], -1)
        output_detections.append(np.concatenate((boxes, scores), -1))
    return output_detections


def non_max_suppression(boxes, probs, overlap_threshold=0.3):
    if len(boxes) == 0:
        return [], []

    if boxes.dtype.kind == "i":
        boxes = boxes.astype("float")

    pick = []

    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]

    area = (x2 - x1) * (y2 - y1)
    idxs = y2

    if probs is not None:
        idxs = probs

    idxs = np.argsort(idxs)

    while len(idxs) > 0:
        last = len(idxs) - 1
        i = idxs[last]
        pick.append(i)

        xx1 = np.maximum(x1[i], x1[idxs[:last]])
        yy1 = np.maximum(y1[i], y1[idxs[:last]])
        xx2 = np.minimum(x2[i], x2[idxs[:last]])
        yy2 = np.minimum(y2[i], y2[idxs[:last]])

        w = np.maximum(0, xx2 - xx1)
        h = np.maximum(0, yy2 - yy1)

        overlap = (w * h) / area[idxs[:last]]
        idxs = np.delete(
            idxs, np.concatenate(([last], np.where(overlap > overlap_threshold)[0]))
        )

    return boxes[pick]


def to_tensor_result(packet):
    return {
        name: np.array(packet.getLayerFp16(name))
        for name in [tensor.name for tensor in packet.getRaw().tensors]
    }



from typing import List

import depthai as dai
import numpy as np
from ahrs.filters import Mahony

from depthai_sdk.classes import IMUPacket
from depthai_sdk.oak_outputs.xout.xout_base import XoutBase, StreamXout


class XoutIMU(XoutBase):
    def __init__(self, imu_xout: StreamXout, fps: int):
        self.imu_out = imu_xout
        self._ahrs = Mahony(frequency=fps)
        self._ahrs.Q = np.array([1, 0, 0, 0], dtype=np.float64)

        super().__init__()
        self.name = 'IMU'

    def xstreams(self) -> List[StreamXout]:
        return [self.imu_out]

    def new_msg(self, name: str, msg: dai.IMUData):
        if name not in self._streams:
            return

        arr = []
        for packet in msg.packets:
            gyro_vals = np.array([packet.gyroscope.z, packet.gyroscope.x, packet.gyroscope.y])
            accelero_vals = np.array([packet.acceleroMeter.z, packet.acceleroMeter.x, packet.acceleroMeter.y])
            self._ahrs.Q = self._ahrs.updateIMU(self._ahrs.Q, gyro_vals, accelero_vals)
            rotation = dai.IMUReportRotationVectorWAcc()
            rotation.i = self._ahrs.Q[0]
            rotation.j = self._ahrs.Q[1]
            rotation.k = self._ahrs.Q[2]
            rotation.real = self._ahrs.Q[3]
            arr.append(IMUPacket(self.get_packet_name(), packet, rotation=rotation))
        return arr



import os
from pathlib import Path
from typing import List, Tuple, Dict, Any, Optional

import depthai as dai

try:
    import cv2
except ImportError:
    cv2 = None

from depthai_sdk.readers.abstract_reader import AbstractReader
from depthai_sdk.components.parser import parse_camera_socket

_videoExt = ['.mjpeg', '.avi', '.mp4', '.h265', '.h264']


class VideoCapReader(AbstractReader):
    """
    Reads stream from mp4, mjpeg, h264, h265
    """

    def __init__(self, path: Path, loop: bool = False) -> None:
        self.videos: Dict[str, Any] = {}
        self._closed = False

        # self.initialFrames: Dict[str, Any] = dict()
        # self.shapes: Dict[str, Tuple[int, int]] = dict()
        # self.readers: Dict[str, cv2.VideoCapture] = dict()
        self._is_looped = loop

        if path.is_file():
            stream_name = path.stem if (path.stem in ['left', 'right']) else 'color'
            self.videos[stream_name] = {
                'reader': cv2.VideoCapture(str(path)),
                'socket': dai.CameraBoardSocket.CAM_A
            }
        else:
            for fileName in os.listdir(str(path)):
                f_name, ext = os.path.splitext(fileName)
                if ext not in _videoExt:
                    continue

                # Check if name of the file starts with left.. right.., or CameraBoardSocket
                if f_name.startswith('CAM_'):
                    # Remove everything after CAM_x
                    f_name = f_name[:5]

                socket = None
                try:
                    socket = parse_camera_socket(f_name)
                except ValueError:
                    # Invalid file name
                    pass

                # TODO: avoid changing stream names, just use socket
                # stream = str(socket)
                # if socket == dai.CameraBoardSocket.CAM_A:
                #     stream = 'color'
                # elif socket == dai.CameraBoardSocket.CAM_B:
                #     stream = 'left'
                # elif socket == dai.CameraBoardSocket.CAM_C:
                #     stream = 'right'
                self.videos[f_name.lower()] = {
                    'reader': cv2.VideoCapture(str(path / fileName)),
                    'socket': socket
                }

        for name, video in self.videos.items():
            ok, f = video['reader'].read()
            video['shape'] = (
                f.shape[1],
                f.shape[0]
            )
            video['is_color'] = len(f.shape) == 3
            video['initialFrame'] = f

    def read(self):
        if self._closed:
            return False
        frames = dict()
        for name, video in self.videos.items():
            if video['initialFrame'] is not None:
                frames[name] = video['initialFrame'].copy()
                video['initialFrame'] = None

            if not video['reader'].isOpened():
                return False

            ok, frame = video['reader'].read()
            if not ok and self._is_looped:
                video['reader'].set(cv2.CAP_PROP_POS_FRAMES, 0)
                ok, frame = video['reader'].read()
            elif not ok:
                return False

            frames[name] = frame

        return frames

    def set_loop(self, loop: bool):
        self._is_looped = loop

    def getStreams(self) -> List[str]:
        return [name for name in self.videos]

    def getShape(self, name: str) -> Tuple[int, int]:
        shape = self.videos[name.lower()]['shape']
        return shape

    def get_socket(self, name: str) -> Optional[dai.CameraBoardSocket]:
        return self.videos[name.lower()]['socket']

    def close(self):
        [r['reader'].release() for _, r in self.videos.items()]
        self._closed = True

    def disableStream(self, name: str):
        if name.lower() in self.videos:
            self.videos.pop(name.lower())

    def get_message_size(self, name: str) -> int:
        video = self.videos[name.lower()]
        return video['shape'][0] * video['shape'][1] * (3 if video['is_color'] else 1)



import time
import warnings
from pathlib import Path
from typing import Dict, Any, Optional, List, Union, Callable

from depthai_sdk.logger import LOGGER
from depthai_sdk import CV2_HAS_GUI_SUPPORT
from depthai_sdk.types import Resolution
from depthai_sdk.visualize.visualizer import Visualizer

try:
    import cv2
except ImportError:
    cv2 = None

import depthai as dai

from depthai_sdk.trigger_action.actions.abstract_action import Action
from depthai_sdk.args_parser import ArgsParser
from depthai_sdk.classes.packet_handlers import (
    BasePacketHandler,
    QueuePacketHandler,
    RosPacketHandler,
    TriggerActionPacketHandler,
    RecordPacketHandler,
    CallbackPacketHandler,
    VisualizePacketHandler
)
# RecordConfig, OutputConfig, SyncConfig, RosStreamConfig, TriggerActionConfig
from depthai_sdk.components.camera_component import CameraComponent
from depthai_sdk.components.component import Component, ComponentOutput
from depthai_sdk.components.imu_component import IMUComponent
from depthai_sdk.components.tof_component import ToFComponent
from depthai_sdk.components.nn_component import NNComponent
from depthai_sdk.components.parser import (
    parse_usb_speed,
    parse_camera_socket,
    get_first_color_cam,
    parse_open_vino_version
)
from depthai_sdk.components.stereo_component import StereoComponent
from depthai_sdk.components.pointcloud_component import PointcloudComponent
from depthai_sdk.record import RecordType, Record
from depthai_sdk.replay import Replay
from depthai_sdk.trigger_action.triggers.abstract_trigger import Trigger
from depthai_sdk.utils import report_crash_dump



class UsbWarning(UserWarning):
    pass


class OakCamera:
    """
    OakCamera improves ease of use when developing apps for OAK devices.

    It abstracts DepthAI API pipeline building, different camera permutations, stream recording/replaying, it adds
    debugging features, does AI model handling, message syncing & visualization, and much more.

    It was designed with interoperability with depthai API in mind.
    """

    def __init__(self,
                 device: Optional[str] = None,
                 usb_speed: Union[None, str, dai.UsbSpeed] = None,  # Auto by default
                 replay: Union[None, str, Path] = None,
                 rotation: int = 0,
                 config: dai.Device.Config = None,
                 args: Union[bool, Dict] = True
                 ):
        """
        Initializes OakCamera

        Args:
            device (str, optional): OAK device we want to connect to, either MxId, IP, or USB port
            usb_speed (str, optional): USB speed we want to use. Defaults to 'auto'.
            replay (str, optional): Replay a depthai-recording - either local path, or from depthai-recordings repo
            rotation (int, optional): Rotate the camera output by this amount of degrees, 0 by default, 90, 180, 270 are supported.
            args (None, bool, Dict): Use user defined arguments when constructing the pipeline
        """

        # User should be able to access these:
        self.replay: Optional[Replay] = None

        self.pipeline = dai.Pipeline()
        self._args: Optional[Dict[str, Any]] = None  # User defined arguments
        self._pipeine_graph = None

        if args:
            if isinstance(args, bool):
                self._args = ArgsParser.parseArgs()
            else:  # Already parsed
                self._args = args
            # Set up the OakCamera
            if self._args.get('recording', None):
                replay = self._args.get('recording', None)
            if self._args.get('deviceId', None):
                device = self._args.get('deviceId', None)
            if self._args.get('usbSpeed', None):
                usb_speed = parse_usb_speed(self._args.get('usbSpeed', None))

            self.config_pipeline(
                xlink_chunk=self._args.get('xlinkChunkSize', None),
                tuning_blob=self._args.get('cameraTuning', None),
                openvino_version=self._args.get('openvinoVersion', None),
            )

        if config is None:
            config = dai.Device.Config()
            config.version = dai.OpenVINO.VERSION_UNIVERSAL
            max_speed = parse_usb_speed(usb_speed)
            if max_speed is not None:
                config.board.usb.maxSpeed = max_speed

        self._init_device(config, device)
        report_crash_dump(self.device)

        # Whether to stop running the OAK camera. Used by oak.running()
        self._stop = False
        self._built = False
        self._polling = []

        self._components: List[Component] = []  # List of components
        self._packet_handlers: List[BasePacketHandler] = []

        self._rotation = rotation
        if replay is not None:
            self.replay = Replay(replay)
            self.replay.initPipeline(self.pipeline)
            LOGGER.info(f'Available streams from recording: {self.replay.getStreams()}')
        self._calibration = self._init_calibration()

    def camera(self,
               source: Union[str, dai.CameraBoardSocket],
               resolution: Optional[Union[
                   str, dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution
               ]] = None,
               fps: Optional[float] = None,
               encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None,
               ) -> CameraComponent:
        """
        Creates Camera component. This abstracts ColorCamera/MonoCamera nodes and supports mocking the camera when
        recording is passed during OakCamera initialization. Mocking the camera will send frames from the host to the
        OAK device (via XLinkIn node).

        Args:
            source (str / dai.CameraBoardSocket): Camera source
            resolution (str/SensorResolution): Sensor resolution of the camera.
            fps (float): Sensor FPS
            encode (bool/str/Profile): Whether we want to enable video encoding (accessible via cameraComponent.out_encoded). If True, it will use MJPEG
        """
        sensor_type = None
        if isinstance(source, str):
            if "," in source:  # For sensors that support multiple
                parts = source.lower().split(',')
                source = parts[0]
                if parts[1] in ["c", "color"]:
                    sensor_type = dai.CameraSensorType.COLOR
                elif parts[1] in ["m", "mono"]:
                    sensor_type = dai.CameraSensorType.MONO
                else:
                    raise Exception(
                        "Please specify sensor type with c/color or m/mono after the ','"
                        " - eg. `cam = oak.create_camera('cama,c')`"
                    )

            if source == 'left':
                source = self._calibration.getStereoLeftCameraId()
            elif source == 'right':
                source = self._calibration.getStereoRightCameraId()
            elif source in ['color', 'rgb']:
                source = get_first_color_cam(self.device)
            else:
                source = parse_camera_socket(source)

            if source in [None, dai.CameraBoardSocket.AUTO]:
                return None  # There's no camera on this socket

        for comp in self._components:
            if isinstance(comp, CameraComponent) and comp._socket == source:
                return comp

        comp = CameraComponent(self.device,
                               self.pipeline,
                               source=source,
                               resolution=resolution,
                               fps=fps,
                               encode=encode,
                               sensor_type=sensor_type,
                               rotation=self._rotation,
                               replay=self.replay,
                               args=self._args)
        self._components.append(comp)
        return comp

    def _init_device(self,
                     config: dai.Device.Config,
                     device_str: Optional[str] = None,
                     ) -> None:

        """
        Connect to the OAK camera
        """
        self.device = None
        if device_str is not None:
            device_info = dai.DeviceInfo(device_str)
        else:
            (found, device_info) = dai.Device.getFirstAvailableDevice()
            if not found:
                raise Exception("No OAK device found to connect to!")

        self.device = dai.Device(
            config=config,
            deviceInfo=device_info,
        )

        # TODO test with usb3 (SUPER speed)
        if config.board.usb.maxSpeed != dai.UsbSpeed.HIGH and self.device.getUsbSpeed() == dai.UsbSpeed.HIGH:
            warnings.warn("Device connected in USB2 mode! This might cause some issues. "
                          "In such case, please try using a (different) USB3 cable, "
                          "or force USB2 mode 'with OakCamera(usb_speed='usb2') as oak:'", UsbWarning)

    def create_camera(self,
                      source: Union[str, dai.CameraBoardSocket],
                      resolution: Optional[Resolution] = None,
                      fps: Optional[float] = None,
                      encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None,
                      ) -> CameraComponent:
        """
        Deprecated, use camera() instead.

        Creates Camera component. This abstracts ColorCamera/MonoCamera nodes and supports mocking the camera when
        recording is passed during OakCamera initialization. Mocking the camera will send frames from the host to the
        OAK device (via XLinkIn node).

        Args:
            source (str / dai.CameraBoardSocket): Camera source
            resolution (str/SensorResolution): Sensor resolution of the camera.
            fps (float): Sensor FPS
            encode (bool/str/Profile): Whether we want to enable video encoding (accessible via cameraComponent.out_encoded). If True, it will use MJPEG
        """
        return self.camera(source, resolution, fps, encode)

    def all_cameras(self,
                    resolution: Optional[Union[
                        str, dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution
                    ]] = None,
                    fps: Optional[float] = None,
                    encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None,
                    ) -> List[CameraComponent]:
        """
        Creates Camera component for each camera sensors on the OAK camera.

        Args:
            resolution (str/SensorResolution): Sensor resolution of the camera.
            fps (float): Sensor FPS
            encode (bool/str/Profile): Whether we want to enable video encoding (accessible via cameraComponent.out_encoded). If True, it will use MJPEG
        """
        components: List[CameraComponent] = []
        # Loop over all available camera sensors
        if self.replay:
            sources = self.replay.getStreams()  # TODO handle in case the stream is not from a camera
        else:
            sources = [cam_sensor.socket for cam_sensor in self.device.getConnectedCameraFeatures()]
        for source in sources:
            comp = CameraComponent(self.device,
                                   self.pipeline,
                                   source=source,
                                   resolution=resolution,
                                   fps=fps,
                                   encode=encode,
                                   rotation=self._rotation,
                                   replay=self.replay,
                                   args=self._args)
            components.append(comp)

        self._components.extend(components)
        return components

    def create_all_cameras(self,
                           resolution: Optional[Union[
                               str, dai.ColorCameraProperties.SensorResolution,
                               dai.MonoCameraProperties.SensorResolution
                           ]] = None,
                           fps: Optional[float] = None,
                           encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None,
                           ) -> List[CameraComponent]:
        """
        Deprecated, use all_cameras() instead.

        Creates Camera component for each camera sensors on the OAK camera.

        Args:
            resolution (str/SensorResolution): Sensor resolution of the camera.
            fps (float): Sensor FPS
            encode (bool/str/Profile): Whether we want to enable video encoding (accessible via cameraComponent.out_encoded). If True, it will use MJPEG
        """
        return self.all_cameras(resolution, fps, encode)

    def create_nn(self,
                  model: Union[str, Dict, Path],
                  input: Union[CameraComponent, NNComponent],
                  nn_type: Optional[str] = None,
                  tracker: bool = False,  # Enable object tracker - only for Object detection models
                  spatial: Union[None, bool, StereoComponent] = None,
                  decode_fn: Optional[Callable] = None,
                  ) -> NNComponent:
        """
        Creates Neural Network component.

        Args:
            model (str / Path): str for SDK supported model or Path to custom model's json/blob
            input (CameraComponent/NNComponent): Input to the model. If NNComponent (detector), it creates 2-stage NN
            nn_type (str): Type of the network (yolo/mobilenet) for on-device NN result decoding (only needed if blob path was specified)
            tracker: Enable object tracker, if model is object detector (yolo/mobilenet)
            spatial: Calculate 3D spatial coordinates, if model is object detector (yolo/mobilenet) and depth stream is available
            decode_fn: Custom decoding function for the model's output
        """
        if spatial and type(spatial) == bool:
            spatial = self.stereo()

        comp = NNComponent(self.device,
                           self.pipeline,
                           model=model,
                           input=input,
                           nn_type=nn_type,
                           tracker=tracker,
                           spatial=spatial,
                           decode_fn=decode_fn,
                           replay=self.replay,
                           args=self._args)
        self._components.append(comp)
        return comp

    def stereo(self,
               resolution: Union[None, str, dai.MonoCameraProperties.SensorResolution] = None,
               fps: Optional[float] = None,
               left: Union[None, dai.Node.Output, CameraComponent] = None,  # Left mono camera
               right: Union[None, dai.Node.Output, CameraComponent] = None,  # Right mono camera
               encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None
               ) -> StereoComponent:
        """
        Create Stereo camera component. If left/right cameras/component aren't specified they will get created internally.

        Args:
            resolution (str/SensorResolution): If monochrome cameras aren't already passed, create them and set specified resolution
            fps (float): If monochrome cameras aren't already passed, create them and set specified FPS
            left (CameraComponent/dai.node.MonoCamera): Pass the camera object (component/node) that will be used for stereo camera.
            right (CameraComponent/dai.node.MonoCamera): Pass the camera object (component/node) that will be used for stereo camera.
            encode (bool/str/Profile): Whether we want to enable video encoding (accessible via StereoComponent.out.encoded). If True, it will use MJPEG.
        """
        if left is None:
            left = self.camera(source="left", resolution=resolution, fps=fps)
        if right is None:
            right = self.camera(source="right", resolution=resolution, fps=fps)

        if right is None or left is None:
            return None

        comp = StereoComponent(self.device,
                               self.pipeline,
                               left=left,
                               right=right,
                               replay=self.replay,
                               args=self._args,
                               encode=encode)
        self._components.append(comp)
        return comp

    def create_stereo(self,
                      resolution: Union[None, str, dai.MonoCameraProperties.SensorResolution] = None,
                      fps: Optional[float] = None,
                      left: Union[None, dai.Node.Output, CameraComponent] = None,  # Left mono camera
                      right: Union[None, dai.Node.Output, CameraComponent] = None,  # Right mono camera
                      name: Optional[str] = None,
                      encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None
                      ) -> StereoComponent:
        """
        Deprecated, use stereo() instead.

        Create Stereo camera component. If left/right cameras/component aren't specified they will get created internally.

        Args:
            resolution (str/SensorResolution): If monochrome cameras aren't already passed, create them and set specified resolution
            fps (float): If monochrome cameras aren't already passed, create them and set specified FPS
            left (CameraComponent/dai.node.MonoCamera): Pass the camera object (component/node) that will be used for stereo camera.
            right (CameraComponent/dai.node.MonoCamera): Pass the camera object (component/node) that will be used for stereo camera.
            encode (bool/str/Profile): Whether we want to enable video encoding (accessible via StereoComponent.out.encoded). If True, it will use h264 codec.
        """
        return self.stereo(resolution, fps, left, right, encode)

    def create_tof(self,
                   source: Union[str, dai.CameraBoardSocket, None] = None,
                   fps: Optional[float] = None,
                   align_to: Optional[CameraComponent] = None) -> ToFComponent:
        """
        Create ToF component.

        Args:
            source (str / dai.CameraBoardSocket): ToF camera source
            fps (float): Sensor FPS
            align_to (CameraComponent): Align ToF to this camera component
        """
        comp = ToFComponent(self.device, self.pipeline, source, align_to, fps)
        self._components.append(comp)
        return comp

    def create_imu(self) -> IMUComponent:
        """
        Create IMU component
        """
        comp = IMUComponent(self.device, self.pipeline)
        self._components.append(comp)
        return comp

    def create_pointcloud(self,
                          depth_input: Union[None, StereoComponent, ToFComponent, dai.node.StereoDepth, dai.Node.Output] = None,
                          colorize: Union[None, CameraComponent, dai.node.MonoCamera, dai.node.ColorCamera, dai.Node.Output, bool] = None,
                          ) -> PointcloudComponent:

        if colorize is None:
            for component in self._components:
                if isinstance(component, CameraComponent):
                    if component.is_color():
                        colorize = component
                        break
                    else:
                        # ColorCamera has priority
                        colorize = component

        comp = PointcloudComponent(
            self.device,
            self.pipeline,
            depth_input=depth_input,
            colorize=colorize,
            replay=self.replay,
            args=self._args,
        )
        self._components.append(comp)
        return comp

    def config_pipeline(self,
                        xlink_chunk: Optional[int] = None,
                        calib: Optional[dai.CalibrationHandler] = None,
                        tuning_blob: Optional[str] = None,
                        openvino_version: Union[None, str, dai.OpenVINO.Version] = None
                        ):
        """
        Configures DepthAI pipeline.
        @param xlink_chunk: Chunk size of XLink messages. 0 can result in lower latency
        @param calib: Calibration data to be uploaded to OAK
        @param tuning_blob: Camera tuning blob
        @param openvino_version: Force specific OpenVINO version
        """
        if xlink_chunk is not None:
            self.pipeline.setXLinkChunkSize(xlink_chunk)
        if calib is not None:
            self.pipeline.setCalibrationData(calib)
        if tuning_blob is not None:
            self.pipeline.setCameraTuningBlobPath(tuning_blob)
        ov_version = parse_open_vino_version(openvino_version)
        if ov_version is not None:
            self.pipeline.setOpenVINOVersion(ov_version)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    def close(self):
        LOGGER.info("Closing OAK camera")
        if self.replay:
            self.replay.close()

        for handler in self._packet_handlers:
            handler.close()

        self.device.close()

    def _new_oak_msg(self, q_name: str, msg):
        if self._stop:
            return
        if q_name in self._new_msg_callbacks:
            for callback in self._new_msg_callbacks[q_name]:
                callback(q_name, msg)

    def start(self, blocking=False):
        """
        Start the application - upload the pipeline to the OAK device.
        Args:
            blocking (bool):  Continuously loop and call oak.poll() until program exits
        """
        self._new_msg_callbacks = {}
        for node in self.pipeline.getAllNodes():
            if isinstance(node, dai.node.XLinkOut):
                self._new_msg_callbacks[node.getStreamName()] = []

        for handler in self._packet_handlers:
            # Setup PacketHandlers. This will:
            # - Initialize all submodules (eg. Recording, Trigger/Actions, Visualizer)
            # - Create XLinkIn nodes for all components/streams
            handler.setup(self.pipeline, self.device, self._new_msg_callbacks)

        # Upload the pipeline to the device and start it
        self.device.startPipeline(self.pipeline)

        for xlink_name in self._new_msg_callbacks:
            try:
                self.device.getOutputQueue(xlink_name, maxSize=1, blocking=False).addCallback(self._new_oak_msg)
            # TODO: make this nicer, have self._new_msg_callbacks know whether it's replay or not
            except Exception as e:
                if self.replay:
                    self.replay._add_callback(xlink_name, self._new_oak_msg)
                else:
                    raise e

        # Append callbacks to be called from main thread
        # self._polling.append()
        if self._pipeine_graph is not None:
            self._pipeine_graph.create_graph(self.pipeline.serializeToJson()['pipeline'], self.device)
            LOGGER.info('Pipeline graph process started')

        # Call on_pipeline_started() for each component
        for comp in self._components:
            comp.on_pipeline_started(self.device)

        if self.replay:
            self.replay.createQueues(self.device)
            self.replay.start()
            # Called from Replay module on each new frame sent to the device.

        # Check if callbacks (sync/non-sync are set)
        if blocking:
            # Constant loop: get messages, call callbacks
            while self.running():
                self.poll()

    def running(self) -> bool:
        """
        Check if camera is running.
        Returns:
            True if camera is running, False otherwise.
        """
        return not self._stop

    def poll(self) -> Optional[int]:
        """
        Poll events; cv2.waitKey, send controls to OAK (if controls are enabled), update, check syncs.

        Returns: key pressed from cv2.waitKey, or None if
        """
        # if self._stop:
        #     return
        if CV2_HAS_GUI_SUPPORT:
            key = cv2.waitKey(1)
            if key == ord('q'):
                self._stop = True
                return key
        else:
            time.sleep(0.001)
            key = -1

        # TODO: check if components have controls enabled and check whether key == `control`
        if self.replay:
            if key == ord(' '):
                self.replay.toggle_pause()

            if self.replay._stop:
                self._stop = True
                return key

        for poll in self._polling:
            poll()  # Poll all callbacks

        if self.device.isClosed():
            self._stop = True
            return None

        return key

    def sync(self, outputs: Union[ComponentOutput, List[ComponentOutput]], callback: Callable, visualize=False):
        raise DeprecationWarning('sync() is deprecated. Use callback() instead.')

    def record(self,
               outputs: Union[ComponentOutput, List[ComponentOutput]],
               path: str,
               record_type: RecordType = RecordType.VIDEO
               ) -> RecordPacketHandler:
        """
        Record component outputs. This handles syncing multiple streams (eg. left, right, color, depth) and saving
        them to the computer in desired format (raw, mp4, mcap, bag..).

        Args:
            outputs (Component/Component output): Component output(s) to be recorded.
            path: Folder path where to save these streams.
            record_type: Record type.
        """
        handler = RecordPacketHandler(outputs, Record(Path(path).resolve(), record_type))
        self._packet_handlers.append(handler)
        return handler

    def show_graph(self):
        """
        Shows DepthAI Pipeline graph, which can be useful when debugging. You must call this BEFORE calling the oak.start()!
        """
        from depthai_pipeline_graph.pipeline_graph import PipelineGraph
        self._pipeine_graph = PipelineGraph()
        self._polling.append(self._pipeine_graph.update)

    def visualize(self,
                  output: Union[List, ComponentOutput, Component],
                  record_path: Optional[str] = None,
                  scale: float = None,
                  fps=False,
                  callback: Callable = None,
                  visualizer: str = 'opencv'
                  ) -> Visualizer:
        """
        Visualize component output(s). This handles output streaming (OAK->host), message syncing, and visualizing.

        Args:
            output (Component/Component output): Component output(s) to be visualized. If component is passed, SDK will visualize its default output (out()).
            record_path: Path where to store the recording (visualization window name gets appended to that path), supported formats: mp4, avi.
            scale: Scale the output window by this factor.
            fps: Whether to show FPS on the output window.
            callback: Instead of showing the frame, pass the Packet to the callback function, where it can be displayed.
            visualizer: Which visualizer to use. Options: 'opencv', 'depthai-viewer', 'robothub'.
        """
        main_thread = False
        visualizer = visualizer.lower()
        if visualizer in ['opencv', 'cv2']:
            from depthai_sdk.visualize.visualizers.opencv_visualizer import OpenCvVisualizer
            vis = OpenCvVisualizer(scale, fps)
            main_thread = True  # OpenCV's imshow() requires to be called from the main thread
        elif visualizer in ['depthai-viewer', 'depthai_viewer', 'viewer', 'depthai']:
            from depthai_sdk.visualize.visualizers.viewer_visualizer import DepthaiViewerVisualizer
            vis = DepthaiViewerVisualizer(scale, fps)
        elif visualizer in ['robothub', 'rh']:
            raise NotImplementedError('Robothub visualizer is not implemented yet')
        else:
            raise ValueError(f"Unknown visualizer: {visualizer}. Options: 'opencv'")

        handler = VisualizePacketHandler(output,
                                         vis,
                                         callback=callback, record_path=record_path,
                                         main_thread=main_thread)
        self._packet_handlers.append(handler)

        if main_thread:
            self._polling.append(handler._poll)

        return vis

    def queue(self, output: Union[ComponentOutput, Component, List], max_size: int = 30) -> QueuePacketHandler:
        """
        Create a queue for the component output(s). This handles output streaming (OAK->Host) and message syncing.

        Args:
            output: Component output(s) to be visualized. If component is passed, SDK will visualize its default output.
            max_size: Maximum queue size for this output.
        """
        handler = QueuePacketHandler(output, max_size)
        self._packet_handlers.append(handler)
        return handler

    def callback(self,
                 output: Union[List, Callable, Component],
                 callback: Callable,
                 main_thread=False
                 ) -> CallbackPacketHandler:
        """
        Create a callback for the component output(s). This handles output streaming (OAK->Host) and message syncing.

        Args:
            output: Component output(s) to be visualized. If component is passed, SDK will visualize its default output.
            callback: Handler function to which the Packet will be sent.
            main_thread: Whether to run the callback in the main thread. If False, it will call the callback in a separate thread, so some functions (eg. cv2.imshow) won't work.
        """
        handler = CallbackPacketHandler(output, callback=callback, main_thread=main_thread)
        if main_thread:
            self._polling.append(handler._poll)
        self._packet_handlers.append(handler)
        return handler

    def ros_stream(self, output: Union[List, ComponentOutput, Component]) -> RosPacketHandler:
        """
        Publish component output(s) to ROS streams.
        """
        handler = RosPacketHandler(output)
        self._packet_handlers.append(handler)
        return handler

    def trigger_action(self, trigger: Trigger, action: Union[Action, Callable]) -> None:
        self._packet_handlers.append(TriggerActionPacketHandler(trigger, action))

    @property
    def sensors(self) -> List[dai.CameraBoardSocket]:
        """
        Returns list of all sensors added to the pipeline.
        """
        return self.device.getConnectedCameraFeatures()

    def _init_calibration(self) -> dai.CalibrationHandler:
        if self.replay:
            calibration = self.pipeline.getCalibrationData()
        else:
            calibration = self.device.readCalibration()
        if calibration is None:
            LOGGER.warning("No calibration data found on the device or in replay")
        return calibration



import time
from typing import Tuple, List, Union
import depthai as dai
import numpy as np
import rclpy
import rclpy.node as node

# from geometry_msgs.msg import Vector3, Quaternion, Pose2D, Point, Transform, TransformStamped
# from std_msgs.msg import Header, ColorRGBA, String
# from visualization_msgs.msg import ImageMarker
from geometry_msgs.msg import Vector3, Quaternion

from sensor_msgs.msg import CompressedImage, Image, Imu  # , PointCloud2, PointField, Imu  # s, PointCloud
from std_msgs.msg import Header
from builtin_interfaces.msg import Time

from depthai_sdk.integrations.ros.imu_interpolation import ImuInterpolation


class DepthAi2Ros2:
    xyz = dict()

    def __init__(self, device: dai.Device) -> None:
        self.start_time = dai.Clock.now()
        self.device = device
        self.imu_packets = []
        self.imu_interpolation = ImuInterpolation()

    def set_header(self, msg, dai_msg: Union[dai.ImgFrame, dai.IMUReport]) -> Header:
        try:
            msg.header.frame_id = str(dai_msg.getSequenceNum())  # ImgFrame
        except:
            msg.header.frame_id = str(dai_msg.sequence)  # IMUReport

        ts = dai_msg.getTimestampDevice() - self.start_time
        # secs / nanosecs
        msg.header.stamp = Time(sec=ts.seconds, nanosec=ts.microseconds * 1000)
        return msg

    def CompressedImage(self, imgFrame: dai.ImgFrame) -> CompressedImage:
        msg = CompressedImage()
        self.set_header(msg, imgFrame)
        msg.format = "jpeg"
        msg.data.frombytes(imgFrame.getData())
        return msg

    def Image(self, imgFrame: dai.ImgFrame) -> Image:
        msg = Image()
        self.set_header(msg, imgFrame)
        msg.height = imgFrame.getHeight()
        msg.width = imgFrame.getWidth()
        msg.step = imgFrame.getWidth()
        msg.is_bigendian = 0

        type = imgFrame.getType()
        TYPE = dai.ImgFrame.Type
        if type == TYPE.RAW16:  # Depth
            msg.encoding = 'mono16'
            msg.step *= 2  # 2 bytes per pixel
            msg.data.frombytes(imgFrame.getData())
        elif type in [TYPE.GRAY8, TYPE.RAW8]:  # Mono frame
            msg.encoding = 'mono8'
            msg.data.frombytes(imgFrame.getData())
        else:
            msg.encoding = 'bgr8'
            msg.data.frombytes(imgFrame.getCvFrame())
        return msg

    # def TfMessage(self,
    #               imgFrame: dai.ImgFrame,
    #               translation: Tuple[float, float, float] = (0., 0., 0.),
    #               rotation: Tuple[float, float, float, float] = (0., 0., 0., 0.)) -> tfMessage:
    #     msg = tfMessage()
    #     tf = TransformStamped()
    #     tf.header = self.header(imgFrame)
    #     tf.child_frame_id = str(imgFrame.getSequenceNum())
    #     tf.transform = Transform(
    #         translation=Vector3(x=translation[0], y=translation[1], z=translation[2]),
    #         rotation=Quaternion(x=rotation[0], y=rotation[1], z=rotation[2], w=rotation[3])
    #     )
    #     msg.transforms.append(tf)
    #     return msg
    #
    # def PointCloud2(self, imgFrame: dai.ImgFrame) -> PointCloud2:
    #     """
    #     Depth frame -> ROS1 PointCloud2 message
    #     """
    #     msg = PointCloud2()
    #     msg.header = self.header(imgFrame)
    #
    #     heigth = str(imgFrame.getHeight())
    #     if heigth not in self.xyz:
    #         self._create_xyz(imgFrame.getWidth(), imgFrame.getHeight())
    #
    #     frame = imgFrame.getCvFrame()
    #     frame = np.expand_dims(np.array(frame), axis=-1)
    #     pcl = self.xyz[heigth] * frame / 1000.0  # To meters
    #
    #     msg.height = imgFrame.getHeight()
    #     msg.width = imgFrame.getWidth()
    #     msg.fields = [
    #         PointField(name="x", offset=0, datatype=PointField.FLOAT32, count=1),
    #         PointField(name="y", offset=4, datatype=PointField.FLOAT32, count=1),
    #         PointField(name="z", offset=8, datatype=PointField.FLOAT32, count=1)
    #     ]
    #     msg.is_bigendian = False
    #     msg.point_step = 12  # 3 * float32 (=4 bytes)
    #     msg.row_step = 12 * imgFrame.getWidth()
    #     msg.data = pcl.tobytes()
    #     msg.is_dense = True
    #     return msg
    #
    # def _create_xyz(self, width, height):
    #     calibData = self.device.readCalibration()
    #     M_right = calibData.getCameraIntrinsics(dai.CameraBoardSocket.RIGHT, dai.Size2f(width, height))
    #     camera_matrix = np.array(M_right).reshape(3, 3)
    #
    #     xs = np.linspace(0, width - 1, width, dtype=np.float32)
    #     ys = np.linspace(0, height - 1, height, dtype=np.float32)
    #
    #     # generate grid by stacking coordinates
    #     base_grid = np.stack(np.meshgrid(xs, ys))  # WxHx2
    #     points_2d = base_grid.transpose(1, 2, 0)  # 1xHxWx2
    #
    #     # unpack coordinates
    #     u_coord: np.array = points_2d[..., 0]
    #     v_coord: np.array = points_2d[..., 1]
    #
    #     # unpack intrinsics
    #     fx: np.array = camera_matrix[0, 0]
    #     fy: np.array = camera_matrix[1, 1]
    #     cx: np.array = camera_matrix[0, 2]
    #     cy: np.array = camera_matrix[1, 2]
    #
    #     # projective
    #     x_coord: np.array = (u_coord - cx) / fx
    #     y_coord: np.array = (v_coord - cy) / fy
    #
    #     xyz = np.stack([x_coord, y_coord], axis=-1)
    #     self.xyz[str(height)] = np.pad(xyz, ((0, 0), (0, 0), (0, 1)), "constant", constant_values=1.0)
    def Imu(self, dai_msg):
        dai_msg: dai.IMUData
        for packet in dai_msg.packets:
            msg = Imu(
                orientation=Quaternion(x=0.0, y=0.0, z=0.0, w=1.0),
                orientation_covariance=np.array([-1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
                angular_velocity=Vector3(x=0.0, y=0.0, z=0.0),
                angular_velocity_covariance=np.zeros(9),
                linear_acceleration=Vector3(x=0.0,y=0.0, z=0.0),
                linear_acceleration_covariance=np.zeros(9)
            )
            report = packet.acceleroMeter or packet.gyroscope or packet.magneticField or packet.rotationVector
            self.set_header(msg, report)
            self.imu_interpolation.Imu(msg, packet)
            # TODO: publish from here directly, so single IMUData can result in more Imu packets?
            return msg




import warnings
from enum import Enum
from typing import Optional, Union, Any, Dict, Tuple, List

import cv2
import depthai as dai
import numpy as np

from depthai_sdk.components.camera_component import CameraComponent, ComponentOutput
from depthai_sdk.components.component import Component
from depthai_sdk.components.parser import parse_median_filter, parse_encode, encoder_profile_to_fourcc
from depthai_sdk.components.stereo_control import StereoControl
from depthai_sdk.components.undistort import _get_mesh
from depthai_sdk.logger import LOGGER
from depthai_sdk.oak_outputs.xout.xout_base import XoutBase, StreamXout
from depthai_sdk.oak_outputs.xout.xout_depth import XoutDisparityDepth
from depthai_sdk.oak_outputs.xout.xout_disparity import XoutDisparity
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames
from depthai_sdk.replay import Replay
from depthai_sdk.visualize.configs import StereoColor
from depthai_sdk.visualize.visualizer_helper import depth_to_disp_factor


class WLSLevel(Enum):
    """WLS filter level"""
    LOW = (1000, 0.8)
    MEDIUM = (6000, 1.5)
    HIGH = (12000, 2.0)


class StereoComponent(Component):

    @property
    def depth(self) -> dai.Node.Output:
        # Depth output from the StereoDepth node.
        return self.node.depth

    @property
    def disparity(self) -> dai.Node.Output:
        # Disparity output from the StereoDepth node.
        return self.node.disparity

    def __init__(self,
                 device: dai.Device,
                 pipeline: dai.Pipeline,
                 left: Union[CameraComponent, dai.node.MonoCamera],  # Left stereo camera
                 right: Union[CameraComponent, dai.node.MonoCamera],  # Right stereo camera
                 replay: Optional[Replay] = None,
                 args: Any = None,
                 encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None):
        """
        Args:
            device (dai.Device): DepthAI device.
            pipeline (dai.Pipeline): DepthAI pipeline.
            left (dai.None.Output / CameraComponent): Left mono camera source. Will get handled by Camera object.
            right (dai.None.Output / CameraComponent): Right mono camera source. Will get handled by Camera object.
            replay (Replay object, optional): Replay object to use for playback.
            args (Any, optional): Use user defined arguments when constructing the pipeline.
            encode (str/bool/Profile, optional): Encode the output stream.
        """
        super().__init__()
        self.out = self.Out(self)

        self.left: Union[None, CameraComponent, dai.node.MonoCamera, dai.node.ColorCamera, dai.Node.Output]
        self.right: Union[None, CameraComponent, dai.node.MonoCamera, dai.node.ColorCamera, dai.Node.Output]

        self._left_stream: dai.Node.Output
        self._right_stream: dai.Node.Output

        self.colormap = None  # for on-device colorization

        self._device = device
        self._replay: Optional[Replay] = replay
        self._args: Dict = args

        self.left = left
        self.right = right

        self.node: dai.node.StereoDepth = pipeline.createStereoDepth()
        self.node.setDefaultProfilePreset(dai.node.StereoDepth.PresetMode.HIGH_DENSITY)

        self._align_component: Optional[CameraComponent] = None
        self.ir_settings = {
            'auto_mode': False,
            'continuous_mode': False,
        }

        # Encoder
        self.encoder = None
        if encode:
            self.encoder = pipeline.createVideoEncoder()
            # MJPEG by default
            self._encoderProfile = parse_encode(encode)

        # Postprocessing options
        self._colorize = None
        self._postprocess_colormap = None

        self.wls_config = {
            'enabled': None,
            'level': None,
            'lambda': None,
            'sigma': None
        }

        # Output config
        self.enable_depth_score = False
        self.validate_calibration = False

        self._undistortion_offset: Optional[int] = None

        if not self._replay:
            # Live stream, check whether we have correct cameras
            if len(device.getCameraSensorNames()) == 1:
                raise Exception('OAK-1 camera does not have Stereo camera pair!')

            if self._get_ir_drivers():
                laser = self._args.get('irDotBrightness', None)
                laser = laser if laser is not None else 800
                if 0 < laser:
                    device.setIrLaserDotProjectorBrightness(laser)
                    LOGGER.info(f'Setting IR laser dot projector brightness to {laser}mA')

                led = self._args.get('irFloodBrightness', None)
                if led is not None:
                    device.setIrFloodLightBrightness(int(led))
                    LOGGER.info(f'Setting IR flood LED brightness to {int(led)}mA')

            input_size = self._get_stream_size(self.left)
            if input_size:
                self.node.setInputResolution(*input_size)

        self._left_stream = self._get_output_stream(self.left)
        self._right_stream = self._get_output_stream(self.right)

        # Check whether input stereo pairs are larger than 1280 pixels in width (limitation of the RVC2/RVC3).
        # If that's the case, create ImageManip to downscale the streams.
        downscale_manips = []
        if isinstance(self.left, CameraComponent):
            # Check whether input size width is larger than 1280
            w, h = self.left.stream_size
            if w > 1280:
                manip = pipeline.create(dai.node.ImageManip)
                new_h = int(h * (1280 / w))
                manip.setResize(1280, new_h)
                LOGGER.info(f'Input frame size to stereo component was {w}x{h}, added downscalling to 1280x{new_h}')
                manip.setMaxOutputFrameSize(1280 * new_h)
                # Stereo works on GRAY8 frames
                manip.setFrameType(dai.ImgFrame.Type.GRAY8)
                self._left_stream.link(manip.inputImage)
                self._left_stream = manip.out
                downscale_manips.append(manip)
        if isinstance(self.right, CameraComponent):
            # Check whether input size width is larger than 1280
            w, h = self.right.stream_size
            if w > 1280:
                manip = pipeline.create(dai.node.ImageManip)
                new_h = int(h * (1280 / w))
                manip.setResize(1280, new_h)
                LOGGER.info(f'Input frame size to stereo component was {w}x{h}, added downscalling to 1280x{new_h}')
                manip.setMaxOutputFrameSize(1280 * new_h)
                # Stereo works on GRAY8 frames
                manip.setFrameType(dai.ImgFrame.Type.GRAY8)
                self._right_stream.link(manip.inputImage)
                self._right_stream = manip.out
                downscale_manips.append(manip)

        if self._replay:  # Replay
            self._replay.initStereoDepth(self.node, left_name=self.left._source, right_name=self.right._source)
        else:
            self._left_stream.link(self.node.left)
            self._right_stream.link(self.node.right)

        if self.encoder:
            try:
                fps = self.left.get_fps()  # CameraComponent
            except AttributeError:
                fps = self.left.getFps()  # MonoCamera

            self.encoder.setDefaultProfilePreset(fps, self._encoderProfile)
            self.node.disparity.link(self.encoder.input)

        self.node.setRectifyEdgeFillColor(0)

        if self._undistortion_offset is not None:
            calib_data = self._replay._calibData if self._replay else device.readCalibration()
            w_frame, h_frame = self._get_stream_size(self.left)
            mapX_left, mapY_left, mapX_right, mapY_right = self._get_maps(w_frame, h_frame, calib_data)
            mesh_l = _get_mesh(mapX_left, mapY_left)
            mesh_r = _get_mesh(mapX_right, mapY_right)
            mesh_left = list(mesh_l.tobytes())
            mesh_right = list(mesh_r.tobytes())
            self.node.loadMeshData(mesh_left, mesh_right)

        if self._args:
            self._config_stereo_args(self._args)

        self.control = StereoControl(device=device)
        self._control_xlink_in = pipeline.create(dai.node.XLinkIn)
        self._control_xlink_in.setStreamName(f"{self.node.id}_inputControl")
        self._control_xlink_in.out.link(self.node.inputConfig)
        # CameraControl message doesn't use any additional data (only metadata)
        self._control_xlink_in.setMaxDataSize(1)

    def on_pipeline_started(self, device: dai.Device):
        if self._control_xlink_in is not None:
            queue = device.getInputQueue(self._control_xlink_in.getStreamName())
            self.control.set_input_queue(queue)

    def _get_output_stream(self, input: Union[
        CameraComponent, dai.node.MonoCamera, dai.node.ColorCamera, dai.Node.Output
    ]) -> dai.Node.Output:
        if isinstance(input, CameraComponent):
            return input.stream
        elif isinstance(input, dai.node.MonoCamera):
            return input.out
        elif isinstance(input, dai.node.ColorCamera):
            return input.video
        elif isinstance(input, dai.Node.Output):
            return input
        else:
            raise ValueError('get_output_stream() accepts either CameraComponent,'
                             'dai.node.MonoCamera, dai.node.ColorCamera, dai.Node.Output!')

    def _get_stream_size(self,
                         input: Union[CameraComponent, dai.node.MonoCamera, dai.node.ColorCamera, dai.Node.Output]) -> \
            Optional[Tuple[int, int]]:
        if isinstance(input, CameraComponent):
            return input.stream_size
        elif isinstance(input, dai.node.MonoCamera):
            return input.getResolutionSize()
        elif isinstance(input, dai.node.ColorCamera):
            return input.getVideoSize()
        else:
            return None

    def config_undistortion(self, M2_offset: int = 0):
        self._undistortion_offset = M2_offset

    def _config_stereo_args(self, args: Dict):
        if not isinstance(args, Dict):
            args = vars(args)  # Namespace -> Dict

        self.config_stereo(
            confidence=args.get('disparityConfidenceThreshold', None),
            median=args.get('stereoMedianSize', None),
            extended=args.get('extendedDisparity', None),
            subpixel=args.get('subpixel', None),
            lr_check=args.get('lrCheck', None),
            sigma=args.get('sigma', None),
            lr_check_threshold=args.get('lrcThreshold', None),
        )

    def config_stereo(self,
                      confidence: Optional[int] = None,
                      align: Optional[CameraComponent] = None,
                      median: Union[None, int, dai.MedianFilter] = None,
                      extended: Optional[bool] = None,
                      subpixel: Optional[bool] = None,
                      lr_check: Optional[bool] = None,
                      sigma: Optional[int] = None,
                      lr_check_threshold: Optional[int] = None,
                      subpixel_bits: Optional[int] = None,
                      ) -> None:
        """
        Configures StereoDepth modes and options.
        """
        if confidence is not None: self.node.initialConfig.setConfidenceThreshold(confidence)
        if align is not None:
            self._align_component = align
            self.node.setDepthAlign(align._socket)
        if median is not None: self.node.setMedianFilter(parse_median_filter(median))
        if extended is not None: self.node.initialConfig.setExtendedDisparity(extended)
        if subpixel is not None: self.node.initialConfig.setSubpixel(subpixel)
        if lr_check is not None: self.node.initialConfig.setLeftRightCheck(lr_check)
        if sigma is not None: self.node.initialConfig.setBilateralFilterSigma(sigma)
        if lr_check_threshold is not None: self.node.initialConfig.setLeftRightCheckThreshold(lr_check_threshold)
        if subpixel_bits is not None: self.node.initialConfig.setSubpixelFractionalBits(subpixel_bits)

    def config_postprocessing(self,
                              colorize: Union[StereoColor, bool] = None,
                              colormap: int = None
                              ) -> None:
        """
        Configures postprocessing options.

        Args:
            colorize: Colorize the disparity map. Can be either a StereoColor enum, string or bool.
            colormap: Colormap to use for colorizing the disparity map.
        """
        if colorize is None:
            self._colorize = StereoColor.GRAY
        elif isinstance(colorize, bool):
            self._colorize = StereoColor.RGB if colorize else StereoColor.GRAY
        elif isinstance(colorize, StereoColor):
            self._colorize = colorize
        elif isinstance(colorize, str):
            self._colorize = StereoColor[colorize.upper()]

        self._postprocess_colormap = colormap

    def config_wls(self,
                   wls_level: Union[WLSLevel, str] = None,
                   wls_lambda: float = None,
                   wls_sigma: float = None
                   ) -> None:
        """
        Configures WLS filter options.

        Args:
            wls_level: WLS filter level. Can be either a WLSLevel enum or string.
            wls_lambda: WLS filter lambda.
            wls_sigma: WLS filter sigma.
        """
        if isinstance(wls_level, WLSLevel):
            wls_level = wls_level
        elif isinstance(wls_level, str):
            wls_level = WLSLevel[wls_level.upper()]

        self.wls_config = {
            'enabled': True if wls_level else False,
            'level': wls_level,
            'lambda': wls_lambda,
            'sigma': wls_sigma,
        }

    def config_output(self,
                      depth_score: bool = None,
                      validate_calibration: bool = None
                      ) -> None:
        """
        Configures output streams.

        Args:
            depth_score: True to include depth score in the output packets.
            validate_calibration: Check if the calibration is valid during the runtime (done on-host) and warn
            the user if it's not. This can be used to detect if the calibration is invalid (e.g. due to temperature drift).
        """
        if depth_score is not None:
            self.enable_depth_score = depth_score
        if validate_calibration is not None:
            self.validate_calibration = validate_calibration

    def set_colormap(self, colormap: dai.Colormap):
        """
        Sets the colormap to use for colorizing the disparity map. Used for on-device postprocessing.
        Works only with `encoded` output.
        Note: This setting can affect the performance.

        Args:
            colormap: Colormap to use for colorizing the disparity map.
        """
        if self.colormap != colormap and self.encoder:
            colormap_manip = self.node.getParentPipeline().create(dai.node.ImageManip)
            colormap_manip.initialConfig.setColormap(colormap, self.node.initialConfig.getMaxDisparity())
            colormap_manip.initialConfig.setFrameType(dai.ImgFrame.Type.NV12)
            if self._align_component:
                h, w = self._align_component.node.getIspSize() \
                    if isinstance(self._align_component.node, dai.node.ColorCamera) \
                    else self._align_component.node.getResolutionSize()
            else:
                h, w = self.left.stream_size
            colormap_manip.setMaxOutputFrameSize(h * w * 3)
            self.node.disparity.link(colormap_manip.inputImage)

            if self.encoder:
                self.node.disparity.unlink(self.encoder.input)
                colormap_manip.out.link(self.encoder.input)
        elif not self.encoder:
            warnings.warn('At the moment, colormap can be used only if encoder is enabled.')

        self.colormap = colormap

    def set_auto_ir(self, auto_mode: bool, continuous_mode: bool = False) -> None:
        """
        Enables/disables auto IR dot projector and flood brightness. Selects the best IR brightness level automatically.
        Can be set to continious mode, which will continuously adjust the IR brightness. Otherwise, it will adjust
        the brightness only once when the device is started.

        Args:
            auto_mode: Enable/disable auto IR.
            continuous_mode: Enable/disable continious mode.
        """
        if self._get_ir_drivers():
            self.ir_settings = {
                'auto_mode': auto_mode,
                'continuous_mode': continuous_mode
            }
            self.set_ir(0, 0)

    def set_ir(self, dot_projector_brightness: int = None, flood_brightness: int = None):
        """
        Sets IR brightness and flood.
        """
        if self._get_ir_drivers():
            if dot_projector_brightness is not None:
                self._device.setIrLaserDotProjectorBrightness(dot_projector_brightness)
            if flood_brightness is not None:
                self._device.setIrFloodLightBrightness(flood_brightness)

    def _get_ir_drivers(self) -> List[Tuple[str, int, int]]:
        """
        Returns a list of IR drivers.
        """
        return self._device.getIrDrivers()

    def _get_disparity_factor(self, device: dai.Device) -> float:
        """
        Calculates the disparity factor used to calculate depth from disparity.
        `depth = disparity_factor / disparity`
        @param device: OAK device
        """
        calib = device.readCalibration()
        baseline = calib.getBaselineDistance(useSpecTranslation=True) * 10  # mm
        intrinsics = calib.getCameraIntrinsics(dai.CameraBoardSocket.RIGHT, self.right.getResolutionSize())
        focal_length = intrinsics[0][0]
        disp_levels = self.node.getMaxDisparity() / 95
        return baseline * focal_length * disp_levels

    def _get_maps(self, width: int, height: int, calib: dai.CalibrationHandler):
        image_size = (width, height)
        M1 = np.array(calib.getCameraIntrinsics(calib.getStereoLeftCameraId(), width, height))
        M2 = np.array(calib.getCameraIntrinsics(calib.getStereoRightCameraId(), width, height))
        d1 = np.array(calib.getDistortionCoefficients(calib.getStereoLeftCameraId()))
        d2 = np.array(calib.getDistortionCoefficients(calib.getStereoRightCameraId()))
        R1 = np.array(calib.getStereoLeftRectificationRotation())
        R2 = np.array(calib.getStereoRightRectificationRotation())

        # increaseOffset = -100 if width == 1152 else -166.67
        """ increaseOffset = 0
        M2_focal = M2.copy()
        M2_focal[0][0] += increaseOffset
        M2_focal[1][1] += increaseOffset
        kScaledL = M2_focal
        kScaledR = kScaledL """

        M2[0][0] += self._undistortion_offset
        M2[1][1] += self._undistortion_offset

        mapX_l, mapY_l = cv2.initUndistortRectifyMap(M1, d1, R1, M2, image_size, cv2.CV_32FC1)
        mapX_r, mapY_r = cv2.initUndistortRectifyMap(M2, d2, R2, M2, image_size, cv2.CV_32FC1)
        return mapX_l, mapY_l, mapX_r, mapY_r

    def get_fourcc(self) -> Optional[str]:
        if self.encoder is None:
            return None
        return encoder_profile_to_fourcc(self._encoderProfile)

    """
    Available outputs (to the host) of this component
    """

    def _aligned_frames(self):
        """
        Create aligned frames output if WLS filter is enabled or colorize is set to RGBD
        """
        aligned_frame = None
        if self.wls_config['enabled'] or self._colorize == StereoColor.RGBD:
            aligned_frame = StreamXout(self._right_stream)
        return aligned_frame

    def _try_get_confidence_map(self):
        confidence_map = None
        if self.enable_depth_score:
            confidence_map = StreamXout(self.node.confidenceMap, name='depth_score')
        return confidence_map

    class Out:
        class DepthOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                return XoutDisparityDepth(
                    device=device,
                    frames=StreamXout(self._comp.depth),
                    dispScaleFactor=depth_to_disp_factor(device, self._comp.node),
                    aligned_frame=self._comp._aligned_frames(),
                    colorize=self._comp._colorize,
                    colormap=self._comp._postprocess_colormap,
                    ir_settings=self._comp.ir_settings,
                    confidence_map=self._comp._try_get_confidence_map()

                ).set_comp_out(self)

        class DisparityOut(ComponentOutput):
            def __call__(self, device: dai.Device, fourcc: Optional[str] = None) -> XoutBase:
                return XoutDisparity(
                    device=device,
                    frames=StreamXout(self._comp.encoder.bitstream) if fourcc else
                    StreamXout(self._comp.disparity),
                    disp_factor=255.0 / self._comp.node.getMaxDisparity(),
                    aligned_frame=self._comp._aligned_frames(),
                    colorize=self._comp._colorize,
                    fourcc=fourcc,
                    colormap=self._comp._postprocess_colormap,
                    wls_config=self._comp.wls_config,
                    ir_settings=self._comp.ir_settings,
                    confidence_map=self._comp._try_get_confidence_map()
                ).set_comp_out(self)

        class RectifiedLeftOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                return XoutFrames(StreamXout(self._comp.node.rectifiedLeft, 'Rectified left')).set_comp_out(self)

        class RectifiedRightOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                return XoutFrames(StreamXout(self._comp.node.rectifiedRight, 'Rectified right')).set_comp_out(self)

        class EncodedOut(DisparityOut):
            def __call__(self, device: dai.Device) -> XoutBase:
                if not self._comp.encoder:
                    raise RuntimeError('Encoder not enabled, cannot output encoded frames')
                if self._comp.wls_config['enabled']:
                    warnings.warn('WLS filter is enabled, but cannot be applied to encoded frames.')

                return super().__call__(device, fourcc=self._comp.get_fourcc())

        def __init__(self, stereo_component: 'StereoComponent'):
            self._comp = stereo_component

            self.depth = self.DepthOut(stereo_component)
            self.rectified_left = self.RectifiedLeftOut(stereo_component)
            self.rectified_right = self.RectifiedRightOut(stereo_component)
            self.disparity = self.DisparityOut(stereo_component)
            self.encoded = self.EncodedOut(stereo_component)
            self.main = self.depth



from abc import ABC
import depthai as dai


class Component(ABC):
    """
    SDK component is used as an abstraction to the current DepthAI API node or group of nodes.    
    """

    def on_pipeline_started(self, device: dai.Device) -> None:
        """
        This function gets called after the pipeline has been started. It is called from the main thread.
        It can be used to eg. initialize XlinkIn queues.
        """
        pass


    # So users can use:
    # packets: Dict[Packet] = q.get()
    # depthPacket = packets['depth']
    # depthPacket = packets[stereoComp]
    def __str__(self):
        return self.out.main.__str__()

    def __hash__(self):
        return self.__str__().__hash__()

    def __eq__(self, other):
        if isinstance(other, Component):
            return str(self) == str(other)
        elif isinstance(other, str):
            return str(self) == other
        else:
            return False

class ComponentOutput(ABC):
    """
    Output of a component
    """
    def __init__(self, component: Component):
        """
        If user hasn't specified component's output name, we will
        generate one in Xout class
        """
        self.name = None
        self._comp = component

    def set_name(self, name: str) -> 'ComponentOutput':
        """
        Name component's output, which will be used for packet names. If not specified, it
        will be generated automatically after pipeline is started (after `oak.start()`) by
        combining all Xout Stream names (eg. "6_out;3_out").
        """
        self.name = name
        return self

    # So users can use:
    # packets: Dict[Packet] = q.get()
    # depthPacket = packets['depth']
    # depthPacket = packets[stereoComp.out.depth]
    def __str__(self):
        return self.name

    def __hash__(self):
        return self.__str__().__hash__()

    def __eq__(self, other):
        if isinstance(other, ComponentOutput):
            return str(self) == str(other)
        elif isinstance(other, str):
            return str(self) == other
        else:
            return False




import logging

__all__ = ['set_logging_level']

LOGGER = logging.getLogger(__name__)
"""The DepthAI SDK logger."""

def _configure_logger():
    """
    Configure the logging module.
    """
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    LOGGER.addHandler(handler)

def set_logging_level(level):
    """
    Set the logging level for the DepthAI SDK logger.
    """
    LOGGER.setLevel(level)


_configure_logger()



from typing import Optional, Dict

import depthai as dai

from depthai_sdk.classes.packets import DisparityDepthPacket
from depthai_sdk.oak_outputs.xout.xout_base import StreamXout
from depthai_sdk.oak_outputs.xout.xout_disparity import XoutDisparity
from depthai_sdk.visualize.configs import StereoColor


class XoutDisparityDepth(XoutDisparity):
    def __init__(self,
                 device: dai.Device,
                 frames: StreamXout,
                 dispScaleFactor: float,
                 aligned_frame: Optional[StreamXout],
                 colorize: StereoColor = None,
                 colormap: int = None,
                 ir_settings: dict = None,
                 confidence_map: StreamXout = None):
        self.name = 'Depth'
        super().__init__(device=device,
                         frames=frames,
                         disp_factor=255 / 95,
                         aligned_frame=aligned_frame,
                         colorize=colorize,
                         colormap=colormap,
                         ir_settings=ir_settings,
                         confidence_map=confidence_map)

        self.disp_scale_factor = dispScaleFactor

    def package(self, msgs: Dict) -> DisparityDepthPacket:
        aligned_frame = msgs[self.aligned_frame.name] if self.aligned_frame else None
        confidence_map = msgs[self.confidence_map.name] if self.confidence_map else None
        return DisparityDepthPacket(
            self.get_packet_name(),
            msgs[self.frames.name],
            colorize=self.colorize,
            colormap=self.colormap,
            aligned_frame=aligned_frame,
            disp_scale_factor=self.disp_scale_factor,
            confidence_map=confidence_map
        )



import math
from abc import ABC, abstractmethod
from collections import defaultdict
from types import SimpleNamespace
from typing import Tuple, List, Union

import depthai as dai
import numpy as np
from depthai import ImgDetection

from depthai_sdk.logger import LOGGER
from depthai_sdk.visualize.bbox import BoundingBox
from depthai_sdk.visualize.configs import VisConfig, BboxStyle, TextPosition


def spatials_text(spatials: dai.Point3f):
    return SimpleNamespace(
        x="X: " + ("{:.1f}m".format(spatials.x / 1000) if not math.isnan(spatials.x) else "--"),
        y="Y: " + ("{:.1f}m".format(spatials.y / 1000) if not math.isnan(spatials.y) else "--"),
        z="Z: " + ("{:.1f}m".format(spatials.z / 1000) if not math.isnan(spatials.z) else "--"),
    )


class GenericObject(ABC):
    """
    Generic object used by visualizer.
    """

    def __init__(self, config=VisConfig(), frame_shape: Tuple[int, ...] = None):
        self.config = config
        self.frame_shape = frame_shape
        self._children: List['GenericObject'] = []

    def set_config(self, config: VisConfig) -> 'GenericObject':
        """
        Set the configuration for the current object.

        Args:
            config: instance of VisConfig.

        Returns:
            self
        """
        self.config = config
        return self

    def set_frame_shape(self, frame_shape: Tuple[int, ...]) -> 'GenericObject':
        """
        Set the incoming frame shape for the current object.

        Args:
            frame_shape: frame shape as a tuple of (height, width, channels).

        Returns:
            self
        """
        self.frame_shape = frame_shape
        return self

    def prepare(self) -> 'GenericObject':
        """
        Prepare necessary data for drawing.

        Returns:
            self
        """
        return self

    @abstractmethod
    def serialize(self) -> dict:
        """
        Serialize the object to dict.
        """
        raise NotImplementedError

    def add_child(self, child: 'GenericObject') -> 'GenericObject':
        """
        Add a child object to the current object.

        Args:
            child: instance derived from GenericObject.

        Returns:
            self
        """
        self._children.append(child.set_config(self.config).set_frame_shape(self.frame_shape).prepare())
        return self

    @property
    def children(self) -> List['GenericObject']:
        """
        Get the children of the current object.

        Returns:
            List of children.
        """
        return self._children


class VisImage(GenericObject):
    def __init__(self, image: np.ndarray, frame_shape: Tuple[int, ...]):
        super().__init__(frame_shape=frame_shape)
        self.image = image

    def prepare(self) -> 'VisImage':
        return self

    def serialize(self):
        return self.image

    def draw(self, frame: np.ndarray) -> None:
        pass


class VisBoundingBox(GenericObject):
    """
    Object that represents a single bounding box.
    """

    def __init__(self,
                 bbox: BoundingBox,
                 label: str,
                 color: Tuple[int, int, int],
                 thickness: int,
                 bbox_style: BboxStyle):
        super().__init__()
        self.bbox = bbox
        self.label = label
        self.color = color
        self.thickness = thickness
        self.bbox_style = bbox_style

    def prepare(self) -> 'GenericObject':
        return self

    def serialize(self) -> dict:
        parent = {
            'type': 'bbox',
            'bbox': self.bbox,
            'label': self.label,
            'bbox_color': self.color,
        }
        if len(self._children) > 0:
            children = [child.serialize() for child in self._children]
            parent['children'] = children

        return parent


class VisDetections(GenericObject):
    """
    Object that represents detections.
    """

    def __init__(self,
                 detections: List[Union[ImgDetection, dai.Tracklet]],
                 normalizer: BoundingBox,
                 label_map: List[Tuple[str, Tuple]] = None,
                 label_color: Tuple[int, int, int] = None,
                 label_background_color: Tuple[int, int, int] = None,
                 label_background_transparency: float = None,
                 spatial_points: List[dai.Point3f] = None,
                 is_spatial=False,
                 parent_bbox: Union[np.ndarray, Tuple[int, int, int, int]] = None):
        """
        Args:
            detections: List of detections.
            normalizer: Normalizer object.
            label_map: List of tuples (label, color).
            spatial_points: List of spatial points. None if not spatial.
            is_spatial: Flag that indicates if the detections are spatial.
            parent_bbox: Bounding box, if there's a detection inside a bounding box.
        """
        super().__init__()
        self.detections = detections
        self.normalizer = normalizer
        self.label_map = label_map
        self.label_color = label_color
        self.label_background_color = label_background_color
        self.label_background_transparency = label_background_transparency
        self.spatial_points = spatial_points
        self.is_spatial = is_spatial
        self.parent_bbox = parent_bbox

        self.bboxes = []
        self.labels = []
        self.colors = []

        try:  # Check if the detections are of type TrackingDetection
            self.detections = [t.srcImgDetection for t in self.detections]
        except AttributeError:
            pass

    def serialize(self) -> dict:
        parent = {
            'type': 'detections',
            'detections': [{
                'bbox': bbox.to_tuple(frame_shape=self.frame_shape) if isinstance(bbox, BoundingBox) else bbox,
                'label': label,
                'color': color,
                'label_color': self.label_color,
                'label_background_color': self.label_background_color,
                'label_background_transparency': self.label_background_transparency
            } for bbox, label, color in list(self.get_detections())]
        }
        if len(self._children) > 0:
            children = [child.serialize() for child in self._children]
            parent['children'] = children

        return parent

    def register_detection(self,
                           bbox: Union[Tuple[int, ...], BoundingBox],
                           label: str,
                           color: Tuple[int, int, int]
                           ) -> None:
        """
        Register a detection.

        Args:
            bbox: Bounding box.
            label: Label.
            color: Color.
        """
        self.bboxes.append(bbox)
        self.labels.append(label)
        self.colors.append(color)

    def prepare(self) -> 'VisDetections':
        detection_config = self.config.detection

        for i, detection in enumerate(self.detections):
            # Get normalized bounding box
            normalized_bbox = self.normalizer.get_relative_bbox(BoundingBox(detection))
            if len(self.frame_shape) < 2:
                LOGGER.debug(f'Visualizer: skipping detection because frame shape is invalid: {self.frame_shape}')
                return self

            # TODO can normalize accept frame shape?

            if self.label_map:
                label, color = self.label_map[detection.label]
            else:
                label, color = str(detection.label), detection_config.color

            if self.is_spatial or self.spatial_points:
                try:
                    spatial_point = detection.spatialCoordinates
                except AttributeError:
                    spatial_point = self.spatial_points[i]

                spatial_coords = spatials_text(spatial_point)

                # Add spatial coordinates
                self.add_child(VisText(f'{spatial_coords.x}\n{spatial_coords.y}\n{spatial_coords.z}',
                                       color=self.label_color or color,
                                       bbox=normalized_bbox,
                                       position=TextPosition.BOTTOM_RIGHT,
                                       background_color=self.label_background_color,
                                       background_transparency=self.label_background_transparency))

            if cv2 and not detection_config.hide_label and len(label) > 0:
                # Place label in the bounding box
                self.add_child(VisText(text=label.capitalize(),
                                       color=self.label_color or color,
                                       bbox=normalized_bbox,
                                       position=detection_config.label_position,
                                       padding=detection_config.label_padding))

            self.register_detection(normalized_bbox, label, color)

        return self

    def get_detections(self) -> List[Tuple[np.ndarray, str, Tuple[int, int, int]]]:
        """
        Get detections.

        Returns:
            List of tuples (bbox, label, color).
        """
        return list(zip(self.bboxes, self.labels, self.colors))


class VisText(GenericObject):
    """
    Object that represents a text.
    """

    def __init__(self,
                 text: str,
                 coords: Tuple[int, int] = None,
                 size: int = None,
                 color: Tuple[int, int, int] = None,
                 thickness: int = None,
                 outline: bool = True,
                 background_color: Tuple[int, int, int] = None,
                 background_transparency: float = 0.5,
                 bbox: Union[np.ndarray, Tuple[int, int, int, int], BoundingBox] = None,
                 position: TextPosition = TextPosition.TOP_LEFT,
                 padding: int = 10):
        """
        If you want to place the text in a bounding box, you can specify the bounding box and the position of the text.
        Please be aware, that in this case the coords must be equal to None, since the coords are calculated based on the
        bounding box and the position.

        .. note::
            `coords` and `bbox` arguments are mutually exclusive. If you specify `coords`, `bbox` will be ignored.

        Args:
            text: Text content.
            coords: Text coordinates.
            size: Font size
            color: Text color.
            thickness: Font thickness.
            outline: Enable outline if set to True, disable otherwise.
            background_color: Background color.
            background_transparency: Background transparency.
            bbox: Bounding box where to place text.
            position: Position w.r.t. to frame (or bbox if is set).
            padding: Padding.
        """
        super().__init__()
        self.text = text
        self.coords = coords
        self.size = size
        self.color = color
        self.thickness = thickness
        self.outline = outline
        self.background_color = background_color
        self.background_transparency = background_transparency
        self.bbox = bbox
        self.position = position
        self.padding = padding

    def serialize(self):
        return {
            'type': 'text',
            'text': self.text,
            'coords': self.coords,
            'color': self.color,
            'thickness': self.thickness,
            'outline': self.outline,
            'background_color': self.background_color,
            'background_transparency': self.background_transparency
        }


class VisTrail(GenericObject):
    """
    Object that represents a trail.
    """

    def __init__(self,
                 tracklets: List[dai.Tracklet],
                 label_map: List[Tuple[str, Tuple]],
                 bbox: BoundingBox):
        """
        Args:
            tracklets: List of tracklets.
            label_map: List of tuples (label, color).
        """
        super().__init__()

        self.tracklets = tracklets
        self.label_map = label_map
        self.bbox = bbox

    def serialize(self):
        parent = {
            'type': 'trail',
            'label_map': self.label_map,
        }
        if len(self.children) > 0:
            children = [c.serialize() for c in self.children]
            parent['children'] = children

        return parent

    def prepare(self) -> 'VisTrail':
        grouped_tracklets = self.groupby_tracklet()
        h, w = self.frame_shape[:2]
        tracking_config = self.config.tracking

        for tracklet_id, tracklets in grouped_tracklets.items():
            color = tracking_config.line_color
            if color is None and self.label_map:
                label, color = self.label_map[tracklets[0].label]
            else:
                label, color = str(tracklets[0].label), tracking_config.line_color

            tracklet_length = 0
            for i in reversed(range(len(tracklets) - 1)):
                # Get current and next detections' centroids
                p1 = self.bbox.get_relative_bbox(BoundingBox(tracklets[i].srcImgDetection)) \
                    .get_centroid().denormalize(self.frame_shape)
                p2 = self.bbox.get_relative_bbox(BoundingBox(tracklets[i + 1].srcImgDetection)) \
                    .get_centroid().denormalize(self.frame_shape)

                if tracking_config.max_length != -1:
                    tracklet_length += np.linalg.norm(np.array(p1) - np.array(p2))
                    if tracklet_length > tracking_config.max_length:
                        break

                thickness = tracking_config.line_thickness
                if tracking_config.fading_tails:
                    thickness = max(1, int(np.ceil(thickness * i / len(tracklets))))

                self.add_child(VisLine(p1, p2,
                                       color=color,
                                       thickness=thickness))

        return self

    def groupby_tracklet(self):
        """
        Group tracklets by tracklet id.

        Returns:
            Dictionary of tracklets grouped by tracklet id.
        """
        grouped = defaultdict(list)

        for tracklet in self.tracklets:
            grouped[tracklet.id].append(tracklet)

        return grouped

    @staticmethod
    def get_rect_centroid(rect: dai.Rect, w, h) -> Tuple[int, int]:
        """
        Get centroid of a rectangle.
        """
        return int(w * (rect.x + rect.width) // 2), int(h * (rect.y + rect.height) // 2)


class VisLine(GenericObject):
    """
    Object that represents a line.
    """

    def __init__(self,
                 pt1: Tuple[int, int],
                 pt2: Tuple[int, int],
                 color: Tuple[int, int, int] = None,
                 thickness: int = None):
        """

        Args:
            pt1: Starting point.
            pt2: Ending point.
            color: Color of the line.
            thickness: Thickness of the line.
        """
        super().__init__()

        self.pt1 = pt1
        self.pt2 = pt2
        self.color = color
        self.thickness = thickness

    def serialize(self):
        parent = {
            'type': 'line',
            'pt1': self.pt1,
            'pt2': self.pt2
        }
        if len(self.children) > 0:
            children = [c.serialize() for c in self.children]
            parent['children'] = children

        return parent

    def prepare(self) -> 'VisLine':
        return self


class VisCircle(GenericObject):
    def __init__(self,
                 coords: Tuple[int, int],
                 radius: int,
                 color: Tuple[int, int, int] = None,
                 thickness: int = None):
        """
        Args:
            coords: Center of the circle.
            radius: Radius of the circle.
            color: Color of the circle.
            thickness: Thickness of the circle.
        """
        super().__init__()

        self.coords = coords
        self.radius = radius
        self.color = color
        self.thickness = thickness

    def prepare(self) -> 'VisCircle':
        return self

    def serialize(self):
        parent = {
            'type': 'circle',
            'center': self.coords,
            'radius': self.radius
        }
        if len(self.children) > 0:
            children = [c.serialize() for c in self.children]
            parent['children'] = children

        return parent


class VisMask(GenericObject):
    def __init__(self, mask: np.ndarray, alpha: float = None):
        super().__init__()
        self.mask = mask
        self.alpha = alpha

    def prepare(self) -> 'VisMask':
        return self

    def serialize(self):
        parent = {
            'type': 'mask',
            'mask': self.mask
        }
        if len(self.children) > 0:
            children = [c.serialize() for c in self.children]
            parent['children'] = children

        return parent


class VisPolygon(GenericObject):
    def __init__(self, polygon):
        super().__init__()
        self.polygon = polygon

    def serialize(self):
        pass

    def prepare(self) -> 'VisPolygon':
        return self

    def draw(self, frame):
        pass




import os
from abc import abstractmethod
from queue import Queue, Empty
from typing import Optional, Callable, List, Union, Dict

import depthai as dai

from depthai_sdk.classes.packets import BasePacket
from depthai_sdk.components.component import Component, ComponentOutput
from depthai_sdk.logger import LOGGER
from depthai_sdk.oak_outputs.fps import FPS
from depthai_sdk.oak_outputs.syncing import TimestampSync
from depthai_sdk.oak_outputs.xout.xout_base import XoutBase, ReplayStream
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames
from depthai_sdk.record import Record
from depthai_sdk.trigger_action.actions.abstract_action import Action
from depthai_sdk.trigger_action.actions.record_action import RecordAction
from depthai_sdk.trigger_action.trigger_action import TriggerAction
from depthai_sdk.trigger_action.triggers.abstract_trigger import Trigger
from depthai_sdk.visualize.visualizer import Visualizer


class BasePacketHandler:
    def __init__(self, main_thread=False):
        self.fps = FPS()
        self.queue = Queue(2) if main_thread else None
        self.outputs: List[ComponentOutput]
        self.sync = None

        self._packet_names = {}  # Check for duplicate packet name, raise error if found (user error)

    @abstractmethod
    def setup(self, pipeline: dai.Pipeline, device: dai.Device, xout_streams: Dict[str, List]):
        raise NotImplementedError()

    def get_fps(self) -> float:
        return self.fps.fps()

    def _new_packet_callback(self, packet: BasePacket):
        """
        Callback from XoutBase. Don't override it. Does FPS counting and calls new_packet().
        """
        if self.sync is not None:
            packet = self.sync.sync(packet.get_timestamp(), packet.name, packet)
            if packet is None:
                return

        self.fps.next_iter()
        if self.queue:
            if self.queue.full():
                self.queue.get()  # Remove oldest packet
            self.queue.put(packet)
        else:
            self.new_packet(packet)

    def configure_syncing(self,
                          enable_sync: bool = True,
                          threshold_ms: int = 17):
        """
        If multiple outputs are used, then PacketHandler can do timestamp syncing of multiple packets
        before calling new_packet().
        Args:
            enable_sync: If True, then syncing is enabled.
            threshold_ms: Maximum time difference between packets in milliseconds.
        """
        if enable_sync:
            if len(self.outputs) < 2:
                LOGGER.error('Syncing requires at least 2 outputs! Skipping syncing.')
                return
            self.sync = TimestampSync(len(self.outputs), threshold_ms)

    def _poll(self):
        """
        Called from main thread.
        """
        if self.queue:
            try:
                packet = self.queue.get_nowait()
                self.new_packet(packet)
            except Empty:
                pass

    @abstractmethod
    def new_packet(self, packet):
        raise NotImplementedError()

    def close(self):
        """
        Used as a cleanup method (eg. close recording), other classes can override it.
        """
        pass

    def _save_outputs(self, output: Union[List, ComponentOutput, Component]):
        if not isinstance(output, List):
            output = [output]

        for i in range(len(output)):
            if isinstance(output[i], Component):
                # Select default (main) output of the component
                output[i] = output[i].out.main

        self.outputs = output

    def _create_xout(self,
                     pipeline: dai.Pipeline,
                     xout: XoutBase,
                     xout_streams: Dict,
                     custom_callback: Callable = None,
                     custom_packet_postfix: str = None):
        # Check for duplicate packet name, raise error if found (user error)
        if custom_packet_postfix:
            xout.set_packet_name_postfix(custom_packet_postfix)

        name = xout.get_packet_name()
        if name in self._packet_names:
            raise ValueError(
                f'User specified duplicate packet name "{name}"! Please specify unique names (or leave empty) for each component output.')
        self._packet_names[name] = True

        # Assign which callback to call when packet is prepared
        xout.new_packet_callback = custom_callback or self._new_packet_callback

        for xstream in xout.xstreams():
            if xstream.name not in xout_streams:
                xout_streams[xstream.name] = []
                if not isinstance(xstream, ReplayStream):
                    xlink = pipeline.createXLinkOut()
                    xlink.setStreamName(xstream.name)
                    xstream.stream.link(xlink.input)
            xout_streams[xstream.name].append(xout.device_msg_callback)


class VisualizePacketHandler(BasePacketHandler):
    def __init__(self,
                 outputs,
                 visualizer: Visualizer,
                 callback: Callable = None,
                 record_path: Optional[str] = None,
                 main_thread: bool = True,
                 ):
        self._save_outputs(outputs)

        if 1 < len(self.outputs) and record_path is not None:
            raise Exception('Recording multiple streams is not supported! '
                            'Call oak.visualize(out, record_path="vid.mp4") for each stream separately')

        self.callback = callback  # Callback that gets called after syncing
        self.visualizer = visualizer
        self.record_path = record_path
        self.recorder = None
        # Main thread: if opencv visualizer, then we need to poll it
        super().__init__(main_thread)

    def setup(self, pipeline: dai.Pipeline, device: dai.Device, xout_streams: Dict[str, List]):
        for output in self.outputs:
            xout: XoutBase = output(device)
            self._create_xout(pipeline, xout, xout_streams)

    def new_packet(self, packet: BasePacket):
        # Create visualizer objects for the visualizer. These objects will then be visualized
        # by the selected visualizer
        packet.prepare_visualizer_objects(self.visualizer)

        if self.callback:
            # Add self.visualizer to packet attributes
            packet.visualizer = self.visualizer
            self.callback(packet)
        else:
            self.visualizer.show(packet)

        if self.recorder:
            self.recorder.write(packet)

    def close(self):
        self.visualizer.close()


class RecordPacketHandler(BasePacketHandler):
    def __init__(self, outputs, recorder: Record):
        self._save_outputs(outputs)
        self.recorder = recorder
        super().__init__()

    def setup(self, pipeline: dai.Pipeline, device: dai.Device, xout_streams: Dict[str, List]):
        xouts: List[XoutFrames] = []
        for output in self.outputs:
            xout = output(device)
            xouts.append(xout)
            self._create_xout(pipeline, xout, xout_streams)

        self.recorder.start(device, xouts)

    def new_packet(self, packet: BasePacket):
        self.recorder.write(packet)

    def close(self):
        self.recorder.close()


class CallbackPacketHandler(BasePacketHandler):
    def __init__(self, outputs, callback: Callable, main_thread=False):
        self._save_outputs(outputs)
        self.callback = callback
        super().__init__(main_thread)

    def setup(self, pipeline: dai.Pipeline, device: dai.Device, xout_streams: Dict[str, List]):
        for output in self.outputs:
            xout = output(device)
            self._create_xout(pipeline, xout, xout_streams)

    def new_packet(self, packet):
        self.callback(packet)


class QueuePacketHandler(BasePacketHandler):
    def __init__(self, outputs, max_size: int):
        super().__init__()
        self._save_outputs(outputs)
        self.queue = Queue(max_size)

    def get_queue(self) -> Queue:
        return self.queue

    def setup(self, pipeline: dai.Pipeline, device: dai.Device, xout_streams: Dict[str, List]):
        for output in self.outputs:
            xout = output(device)
            self._create_xout(pipeline, xout, xout_streams)

    def configure_syncing(self,
                          enable_sync: bool = True,
                          threshold_ms: int = 17) -> 'QueuePacketHandler':
        """
        If multiple outputs are used, then PacketHandler can do timestamp syncing of multiple packets
        before calling new_packet().
        Args:
            enable_sync: If True, then syncing is enabled.
            threshold_ms: Maximum time difference between packets in milliseconds.
        """
        super().configure_syncing(enable_sync, threshold_ms)
        return self

    def new_packet(self, packet):
        # It won't be called, we just added this function to satisfy the abstract class
        pass


class RosPacketHandler(BasePacketHandler):
    def __init__(self, outputs):
        super().__init__()
        self._save_outputs(outputs)

        envs = os.environ
        if 'ROS_VERSION' not in envs:
            raise Exception('ROS installation not found! Please install or source the ROS you would like to use.')

        version = envs['ROS_VERSION']
        if version == '1':
            raise Exception('ROS1 publsihing is not yet supported!')
            from depthai_sdk.integrations.ros.ros1_streaming import Ros1Streaming
            self.ros = Ros1Streaming()
        elif version == '2':
            from depthai_sdk.integrations.ros.ros2_streaming import Ros2Streaming
            self.ros = Ros2Streaming()
        else:
            raise Exception(f"ROS version '{version}' not recognized! Should be either '1' or '2'")

    def setup(self, pipeline: dai.Pipeline, device: dai.Device, xout_streams: Dict[str, List]):
        xouts = []
        for output in self.outputs:
            xout = output(device)
            self._create_xout(pipeline, xout, xout_streams)
            xouts.append(xout)

        self.ros.update(device, xouts)

    def new_packet(self, packet):
        # self.ros.new_msg(name, msg)
        # TODO: implement
        pass

    # def is_ros1(self) -> bool:
    #     try:
    #         import rospy
    #         return True
    #     except:
    #         return False
    #
    # def is_ros2(self):
    #     try:
    #         import rclpy
    #         return True
    #     except:
    #         return False


class TriggerActionPacketHandler(BasePacketHandler):
    def __init__(self, trigger: Trigger, action: Union[Callable, Action]):
        super().__init__()
        self.trigger = trigger
        self.action = Action(None, action) if isinstance(action, Callable) else action
        self.controller = TriggerAction(self.trigger, self.action)

    def setup(self, pipeline: dai.Pipeline, device: dai.Device, xout_streams: Dict[str, List]):
        trigger_xout: XoutBase = self.trigger.input(device)
        self._create_xout(pipeline=pipeline,
                          xout=trigger_xout,
                          xout_streams=xout_streams,
                          custom_callback=self.controller.new_packet_trigger,
                          custom_packet_postfix='trigger')

        if isinstance(self.action, Callable):
            self._save_outputs([trigger_xout])
            return

        action_xouts = []
        if self.action.inputs:
            for output in self.action.inputs:
                xout: XoutBase = output(device)
                xout.new_packet_callback = self.controller.new_packet_action
                self._create_xout(pipeline=pipeline,
                                  xout=xout,
                                  xout_streams=xout_streams,
                                  custom_callback=self.controller.new_packet_action,
                                  custom_packet_postfix='action')
                action_xouts.append(xout)

        if isinstance(self.action, RecordAction):
            self.action.setup(device, action_xouts)  # creates writers for VideoRecorder()

        self._save_outputs([trigger_xout] + action_xouts)

    def new_packet(self, packet):
        pass


class StreamPacketHandler(BasePacketHandler):
    """
    TODO. API:
    oak.stream_rtsp([color, left, right], port=8888)
    oak.stream_webrtc(color, port=8881)

    Creates a server and just sends forward
    the frames. Doesn't use any queues.

    """
    pass



from typing import Optional, Sequence, Tuple, Union

import cv2
import numpy as np

from depthai_sdk.visualize.bbox import BoundingBox
from depthai_sdk.visualize.configs import VisConfig
from depthai_sdk.visualize.objects import VisText


class OpenCvTextVis:
    def __init__(self, text: VisText, config: VisConfig):
        self.text = text
        self.config = config

    def draw_text(self, frame: np.ndarray):
        obj = self.text

        self.prepare(frame.shape)

        text_config = self.config.text

        # Extract shape of the bbox if exists
        if obj.bbox is not None:
            # shape = self.bbox[2] - self.bbox[0], self.bbox[3] - self.bbox[1]
            tl, br = obj.bbox.denormalize(frame.shape)
            shape = br[0] - tl[0], br[1] - tl[1]
        else:
            shape = frame.shape[:2]

        font_scale = obj.size or text_config.font_scale
        if obj.size is None and text_config.auto_scale:
            font_scale = self.get_text_scale(shape, obj.bbox)

        # Calculate font thickness
        font_thickness = max(1, int(font_scale * 2)) \
            if text_config.auto_scale else obj.thickness or text_config.font_thickness

        dy = cv2.getTextSize(obj.text, text_config.font_face, font_scale, font_thickness)[0][1] + 10

        for line in obj.text.splitlines():
            y = obj.coords[1]

            if obj.outline:
                # Background
                cv2.putText(img=frame,
                            text=line,
                            org=obj.coords,
                            fontFace=text_config.font_face,
                            fontScale=font_scale,
                            color=text_config.background_color,
                            thickness=font_thickness + 1,
                            lineType=text_config.line_type)

            # Front text
            cv2.putText(img=frame,
                        text=line,
                        org=obj.coords,
                        fontFace=text_config.font_face,
                        fontScale=font_scale,
                        color=obj.color or text_config.font_color,
                        thickness=font_thickness,
                        lineType=text_config.line_type)

            obj.coords = (obj.coords[0], y + dy)

    def get_relative_position(self, obj: VisText, frame_shape) -> Tuple[int, int]:
        """
        Get relative position of the text w.r.t. the bounding box.
        If bbox is None,the position is relative to the frame.
        """
        if obj.bbox is None:
            obj.bbox = BoundingBox()
        text_config = self.config.text

        tl, br = obj.bbox.denormalize(frame_shape)
        shape = br[0] - tl[0], br[1] - tl[1]

        bbox_arr = obj.bbox.to_tuple(frame_shape)

        font_scale = obj.size or text_config.font_scale
        if obj.size is None and text_config.auto_scale:
            self.get_text_scale(shape, bbox_arr)

        text_width, text_height = 0, 0
        for text in obj.text.splitlines():
            text_size = cv2.getTextSize(text=text,
                                        fontFace=text_config.font_face,
                                        fontScale=font_scale,
                                        thickness=text_config.font_thickness)[0]
            text_width = max(text_width, text_size[0])
            text_height += text_size[1]

        x, y = bbox_arr[0], bbox_arr[1]

        y_pos = obj.position.value % 10
        if y_pos == 0:  # Y top
            y = bbox_arr[1] + text_height + obj.padding
        elif y_pos == 1:  # Y mid
            y = (bbox_arr[1] + bbox_arr[3]) // 2 + text_height // 2
        elif y_pos == 2:  # Y bottom
            y = bbox_arr[3] - text_height - obj.padding

        x_pos = obj.position.value // 10
        if x_pos == 0:  # X Left
            x = bbox_arr[0] + obj.padding
        elif x_pos == 1:  # X mid
            x = (bbox_arr[0] + bbox_arr[2]) // 2 - text_width // 2
        elif x_pos == 2:  # X right
            x = bbox_arr[2] - text_width - obj.padding

        return x, y

    def prepare(self, frame_shape):
        # TODO: in the future, we can stop support for passing pixel-space bbox to the 
        # visualizer.
        if isinstance(self.text.bbox, (Sequence, np.ndarray)) and type(self.text.bbox[0]) == int:
            # Convert to BoundingBox. Divide by self.frame_shape and load into the BoundingBox
            bbox = list(self.text.bbox)
            bbox[0] /= frame_shape[1]
            bbox[1] /= frame_shape[0]
            bbox[2] /= frame_shape[1]
            bbox[3] /= frame_shape[0]
            self.text.bbox = BoundingBox(bbox)

        self.text.coords = self.text.coords or self.get_relative_position(self.text, frame_shape)

    def get_text_scale(self,
                       frame_shape: Union[np.ndarray, Tuple[int, ...]],
                       bbox: Optional[BoundingBox] = None
                       ) -> float:
        return min(1.0, min(frame_shape) / (1000 if bbox is None else 200))



import threading
from datetime import timedelta
from typing import Dict, Any, Optional


class SequenceNumSync:
    """
        self.msgs = {seqNum: {name: message}}
        Example:

        self.msgs = {
            '1': {
                'rgb': dai.Frame(),
                'dets': dai.ImgDetections(),
            ],
            '2': {
                'rgb': dai.Frame(),
                'dets': dai.ImgDetections(),
            }
        }
        """

    def __init__(self, stream_num: int):
        self.msgs: Dict[str, Dict[str, Any]] = dict()
        self.stream_num: int = stream_num
        self.lock = threading.Lock()

    def sync(self, seq_num: int, name: str, msg) -> Optional[Dict]:
        seq_num = str(seq_num)

        with self.lock:
            if seq_num not in self.msgs:
                self.msgs[seq_num] = dict()

            self.msgs[seq_num][name] = (msg)

            if self.stream_num == len(self.msgs[seq_num]):
                # We have sequence num synced frames!
                ret = self.msgs[seq_num]

                # Remove previous msgs
                new_msgs = {}
                for name, msg in self.msgs.items():
                    if int(name) > int(seq_num):
                        new_msgs[name] = msg
                self.msgs = new_msgs

                return ret

        return None


class TimestampSync:
    def __init__(self, stream_num: int, ms_threshold: int):
        self.msgs: Dict[str, Any] = dict()
        self.stream_num: int = stream_num
        self.ms_threshold = ms_threshold

    def sync(self, timestamp, name: str, msg):
        if name not in self.msgs:
            self.msgs[name] = []

        self.msgs[name].append((timestamp, msg))

        synced = {}
        for name, arr in self.msgs.items():
            # Go through all stored messages and calculate the time difference to the target msg.
            # Then sort these self.msgs to find a msg that's closest to the target time, and check
            # whether it's below 17ms which is considered in-sync.
            diffs = []
            for i, (msg_ts, msg) in enumerate(arr):
                diffs.append(abs(msg_ts - timestamp))
            if len(diffs) == 0:
                break
            diffs_sorted = diffs.copy()
            diffs_sorted.sort()
            dif = diffs_sorted[0]

            if dif < timedelta(milliseconds=self.ms_threshold):
                synced[name] = diffs.index(dif)

        if len(synced) == self.stream_num:  # We have all synced streams
            # Remove older self.msgs
            for name, i in synced.items():
                self.msgs[name] = self.msgs[name][i:]
            ret = {}
            for name, arr in self.msgs.items():
                ts, synced_msg = arr.pop(0)
                ret[name] = synced_msg
            return ret
        return None



from typing import Optional

import cv2
import numpy as np

from depthai_sdk.classes.packets import DisparityPacket, FramePacket
from depthai_sdk.logger import LOGGER
from depthai_sdk.visualize.configs import TextPosition
from depthai_sdk.visualize.objects import (
    VisBoundingBox,
    VisCircle,
    VisDetections,
    VisLine,
    VisMask,
    VisText,
    VisTrail,
)
from depthai_sdk.visualize.visualizer import Visualizer
from depthai_sdk.visualize.visualizer_helper import draw_stylized_bbox, draw_bbox
from depthai_sdk.visualize.visualizers.opencv_text import OpenCvTextVis


class OpenCvVisualizer(Visualizer):
    def draw(self, frame: np.ndarray) -> Optional[np.ndarray]:
        """
        Draw all objects on the frame if the platform is PC. Otherwise, serialize the objects
        and communicate with the RobotHub application.

        Args:
            frame: The frame to draw on.

        Returns:
            np.ndarray if the platform is PC, None otherwise.
        """
        # Draw overlays
        for obj in self.objects:
            if type(obj) == VisBoundingBox:
                draw_stylized_bbox(frame, obj=obj)
            elif type(obj) == VisDetections:
                for bbox, _, color in obj.get_detections():
                    tl, br = bbox.denormalize(frame.shape)
                    draw_bbox(
                        img=frame,
                        pt1=tl,
                        pt2=br,
                        color=color,
                        thickness=self.config.detection.thickness,
                        r=self.config.detection.radius,
                        line_width=self.config.detection.line_width,
                        line_height=self.config.detection.line_height,
                        alpha=self.config.detection.alpha,
                    )
            elif type(obj) == VisText:
                OpenCvTextVis(obj, self.config).draw_text(frame)
            elif type(obj) == VisTrail:
                obj = obj.prepare()
                # Children: VisLine
                self.objects.extend(obj.children)
            elif type(obj) == VisLine:
                cv2.line(frame,
                         obj.pt1, obj.pt2,
                         obj.color or self.config.tracking.line_color,
                         obj.thickness or self.config.tracking.line_thickness,
                         self.config.tracking.line_type)
            elif type(obj) == VisCircle:
                circle_config = self.config.circle
                cv2.circle(frame,
                           obj.coords,
                           obj.radius,
                           obj.color or circle_config.color,
                           obj.thickness or circle_config.thickness,
                           circle_config.line_type)
            elif type(obj) == VisMask:
                cv2.addWeighted(frame, 1 - obj.alpha, obj.mask, obj.alpha, 0, frame)

        self.reset()
        return frame

    def show(self, packet) -> None:
        if self.config.output.show_fps:
            fps = self.fps.get_fps(packet.name)
            self.add_text(text=f'FPS: {fps:.1f}', position=TextPosition.TOP_LEFT)

        if isinstance(packet, DisparityPacket):
            frame = packet.get_colorized_frame(self)
        elif isinstance(packet, FramePacket):
            frame = packet.decode()
        else:
            LOGGER.warning(f'Unknown packet type: {type(packet)}')
            return

        if frame is not None:
            drawn_frame = self.draw(frame)
            if self.config.output.img_scale:
                drawn_frame = cv2.resize(drawn_frame,
                                         None,
                                         fx=self.config.output.img_scale,
                                         fy=self.config.output.img_scale)
            cv2.imshow(packet.name, drawn_frame)

    def close(self):
        cv2.destroyAllWindows()



CV2_HAS_GUI_SUPPORT = False

try:
    import cv2
    import re

    build_info = cv2.getBuildInformation()
    gui_support_regex = re.compile(r'GUI: +([A-Z]+)')
    gui_support_match = gui_support_regex.search(build_info)
    if gui_support_match:
        gui_support = gui_support_match.group(1)
        if gui_support.upper() != 'NONE':
            CV2_HAS_GUI_SUPPORT = True
except ImportError:
    pass



import numpy as np
import depthai as dai
from typing import Tuple
from depthai_sdk.classes.nn_results import ImgLandmarks

THRESHOLD = 0.5

def decode(data: dai.NNData) -> ImgLandmarks:
    # TODO: Use standarized recognition model
    score = np.array(data.getLayerFp16('conv2d_31'))
    score = 1 / (1 + np.exp(-score[0]))  # sigmoid on score
    if score < THRESHOLD:
        return ImgLandmarks(data)

    ldms = np.array(data.getLayerFp16('conv2d_210')).reshape((468, 3))
    landmarks = []

    all_colors = np.array([ldm[2] for ldm in ldms])
    colors = generate_colors_from_z(all_colors)

    for ldm in ldms:
        # colors.append(calculate_color(ldm[2], minz, maxz))
        # colors.append((0, 0, int(ldm[2]) * 5 + 100))
        landmarks.append((ldm[0] / 192, ldm[1] / 192))
    return ImgLandmarks(data, landmarks=landmarks, colors=colors)

def generate_colors_from_z(z_values):
    """
    Generates BGR colors based on normalized Z-values.

    Parameters:
    z_values (numpy.array): Array of Z-values.

    Returns:
    List[Tuple]: List of BGR color tuples.
    """
    def normalize_z_values(z_values, minZ, maxZ):
        return (z_values - minZ) / (maxZ - minZ)

    def map_to_color(normalized_z_values):
        return [(255 - int((1 - value) * 255), 0, 255 - int(value * 255)) for value in normalized_z_values]
    minZ = min(z_values)
    maxZ = max(z_values)
    normalized_z_values = normalize_z_values(z_values, minZ, maxZ)
    return map_to_color(normalized_z_values)


from typing import Dict, Any, Optional, List, Union, Callable
import depthai as dai
from pathlib import Path
import threading
import logging
import time

from depthai_sdk.oak_camera import OakCamera
from depthai_sdk.record import RecordType, Record

from depthai_sdk.components.camera_component import CameraComponent
from depthai_sdk.components.component import Component
from depthai_sdk.components.imu_component import IMUComponent
from depthai_sdk.components.nn_component import NNComponent
from depthai_sdk.components.stereo_component import StereoComponent
from depthai_sdk.components.pointcloud_component import PointcloudComponent
from depthai_sdk.visualize.visualizer import Visualizer

from depthai_sdk.classes.packet_handlers import (
    BasePacketHandler,
    QueuePacketHandler,
    RosPacketHandler,
    TriggerActionPacketHandler,
    RecordPacketHandler,
    CallbackPacketHandler,
    VisualizePacketHandler
)

def flatten(lst):
    flat_list = []
    for item in lst:
        if isinstance(item, list):
            flat_list.extend(flatten(item))
        else:
            flat_list.append(item)
    return flat_list

class OakCameras:
    def __init__(self,
                 devices: Optional[list] = None,
                 usb_speed: Union[None, str, dai.UsbSpeed] = None,  # Auto by default
                 rotation: int = 0,
                 args: Union[bool, Dict] = True):

        if devices is None:
            device_infos = dai.Device.getAllAvailableDevices()
            logging.debug(f'Found {len(device_infos)} devices')
            devices = [device_info.getMxId() for device_info in device_infos]

        if len(devices) == 0:
            raise RuntimeError('No OAK devices found!')

        self.devices: List[OakCamera] = []

        threads = []
        for device in devices:
            # If user already passed OakCamera
            if type(device) is OakCamera:
                self.devices.append(device)
                continue

            def connect(device: str, devices: List[OakCamera]):
                try:
                    logging.info(f'Connecting to {device}')
                    devices.append(OakCamera(device, usb_speed=usb_speed, rotation=rotation, args=args))
                except Exception as e:
                    logging.error(f'Failed to connect to {device}: {e}')

            time.sleep(1) # Currently required due to XLink race issues
            thread = threading.Thread(target=connect, args=(device, self.devices))
            thread.start()
            threads.append(thread)

        for t in threads:
            t.join() # Wait for all threads to finish

        if len(self.devices) == 0:
            raise RuntimeError('No OAK devices connected!')

        self._packet_handlers = []
        self._polling = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        for device in self.devices:
            device.close()

    def camera(self,
            source: Union[str, dai.CameraBoardSocket],
            resolution: Optional[Union[
                str, dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution
                ]] = None,
            fps: Optional[float] = None,
            encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None,
            name: Optional[str] = None,
            ) -> List[CameraComponent]:
        return [device.camera(source, resolution, fps, encode, name) for device in self.devices]

    def all_cameras(self,
                resolution: Optional[Union[
                    str, dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution
                    ]] = None,
                fps: Optional[float] = None,
                encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None,
                ) -> List[CameraComponent]:
        cams = []
        for device in self.devices:
            cams.extend(device.all_cameras(resolution, fps, encode))
        return cams

    def create_nn(self,
                  model: Union[str, Dict, Path],
                  inputs: List[Union[CameraComponent, NNComponent]],
                  nn_type: Optional[str] = None,
                  tracker: bool = False,  # Enable object tracker - only for Object detection models
                  spatials: Union[None, bool, List[StereoComponent]] = None,
                  decode_fn: Optional[Callable] = None,
                  name: Optional[str] = None
                  ) -> List[NNComponent]:
        if len(inputs) != len(self.devices):
            raise RuntimeError(f'Number of inputs ({len(inputs)}) must match number of devices ({len(self.devices)})')

        nns = []
        if type(spatials) is not List:
            spatials = [spatials] * len(self.devices)

        for input, device, spatial in zip(inputs, self.devices, spatials):
            nns.append(device.create_nn(model, input, nn_type, tracker, spatial, decode_fn, name))

        return nns

    def stereo(self,
               resolution: Union[None, str, dai.MonoCameraProperties.SensorResolution] = None,
               fps: Optional[float] = None,
               lefts: Union[None, List[CameraComponent]] = None,  # Left mono camera
               rights: Union[None, List[CameraComponent]] = None,  # Right mono camera
               name: Optional[str] = None,
               encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None
               ) -> List[StereoComponent]:

        if lefts is None and rights is None:
            return [device.stereo(resolution, fps, lefts, rights, name, encode) for device in self.devices]

        stereo = []
        for device, left, right in zip(self.devices, lefts, rights):
            stereo.append(device.stereo(resolution, fps, left, right, name, encode))

        return stereo

    def create_imu(self) -> List[IMUComponent]:
        return [device.create_imu() for device in self.devices]

    def start(self, blocking=False):

        


        for device in self.devices:
            device.start(blocking=False)

        stop = False
        if blocking:
            while not stop:
                for device in self.devices:
                    stop = stop or not device.running()
                    device.poll()

    def poll(self) -> Optional[int]:
        for poll in self._polling:
            poll()

        key = None
        for device in self.devices:
            key = key or device.poll()
        return key

    def record(self,
            outputs: Union[Callable, List[Callable]],
            path: str,
            record_type: RecordType = RecordType.VIDEO) -> RecordPacketHandler:

        outputs = flatten(outputs)
        print('outputs',outputs)

        handler = RecordPacketHandler(outputs, Record(Path(path).resolve(), record_type))
        self._packet_handlers.append(handler)
        return handler

    def visualize(self,
                  outputs: Union[List, Callable, Component],
                  record_path: Optional[str] = None,
                  scale: float = None,
                  fps=False,
                  callback: Callable = None,
                  visualizer: str = 'opencv'
                  ) -> Visualizer:
        """
        Visualize component output(s). This handles output streaming (OAK->host), message syncing, and visualizing.
        Args:
            output (Component/Component output): Component output(s) to be visualized. If component is passed, SDK will visualize its default output (out())
            record_path: Path where to store the recording (visualization window name gets appended to that path), supported formats: mp4, avi
            scale: Scale the output window by this factor
            fps: Whether to show FPS on the output window
            callback: Instead of showing the frame, pass the Packet to the callback function, where it can be displayed
        """
        if type(outputs) is list:
            outputs = flatten(outputs)

        main_thread = False
        visualizer = visualizer.lower()
        if visualizer in ['opencv', 'cv2']:
            from depthai_sdk.visualize.visualizers.opencv_visualizer import OpenCvVisualizer
            vis = OpenCvVisualizer(scale, fps)
            main_thread=True # OpenCV's imshow() requires to be called from the main thread
        elif visualizer in ['depthai-viewer', 'depthai_viewer', 'viewer', 'depthai']:
            from depthai_sdk.visualize.visualizers.viewer_visualizer import DepthaiViewerVisualizer
            vis = DepthaiViewerVisualizer(scale, fps)
        elif visualizer in ['robothub', 'rh']:
            raise NotImplementedError('Robothub visualizer is not implemented yet')
        else:
            raise ValueError(f"Unknown visualizer: {visualizer}. Options: 'opencv'")

        handler = VisualizePacketHandler(outputs, vis, callback=callback, record_path=record_path, main_thread=main_thread)
        self._packet_handlers.append(handler)

        if main_thread:
            self._polling.append(handler._poll)

        return vis









from typing import List, Optional

import depthai as dai
import numpy as np

from depthai_sdk.classes.packets import PointcloudPacket
from depthai_sdk.components.pointcloud_helper import create_xyz
from depthai_sdk.oak_outputs.xout.xout_base import StreamXout
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames

try:
    import cv2
except ImportError:
    cv2 = None


class XoutPointcloud(XoutFrames):
    def __init__(self,
                 device: dai.Device,
                 depth_frames: StreamXout,
                 color_frames: Optional[StreamXout] = None):
        self.color_frames = color_frames
        XoutFrames.__init__(self, frames=depth_frames)
        self.name = 'Pointcloud'
        self.device = device
        self.xyz = None

        self.msgs = dict()

    def xstreams(self) -> List[StreamXout]:
        if self.color_frames is not None:
            return [self.frames, self.color_frames]
        return [self.frames]

    def new_msg(self, name: str, msg: dai.Buffer):
        if name not in self._streams:
            return  # From Replay modules. TODO: better handling?

        # TODO: what if msg doesn't have sequence num?
        seq = str(msg.getSequenceNum())

        if seq not in self.msgs:
            self.msgs[seq] = dict()

        if name == self.frames.name:
            self.msgs[seq][name] = msg
        elif name == self.color_frames.name:
            self.msgs[seq][name] = msg
        else:
            raise ValueError('Message from unknown stream name received by XOutPointcloud!')

        if len(self.msgs[seq]) == len(self.xstreams()):
            # Frames synced!
            depth_frame: dai.ImgFrame = self.msgs[seq][self.frames.name]

            color_frame = None
            if self.color_frames is not None:
                color_frame: dai.ImgFrame = self.msgs[seq][self.color_frames.name]

            if self.xyz is None:
                self.xyz = create_xyz(self.device, depth_frame.getWidth(), depth_frame.getHeight())

            pcl = self.xyz * np.expand_dims(np.array(depth_frame.getFrame()), axis=-1)

            # TODO: postprocessing
            # Cleanup
            new_msgs = {}
            for name, msg in self.msgs.items():
                if int(name) > int(seq):
                    new_msgs[name] = msg
            self.msgs = new_msgs

            return PointcloudPacket(
                self.get_packet_name(),
                pcl,
                depth_map=depth_frame,
                colorize_frame=color_frame
            )



import threading
from typing import List, Union, Dict, Any, Optional, Tuple

import depthai as dai

from depthai_sdk.classes import Detections, ImgLandmarks, SemanticSegmentation
from depthai_sdk.classes.enum import ResizeMode
from depthai_sdk.classes.packets import (
    Detection,
    DetectionPacket,
    ImgLandmarksPacket,
    NnOutputPacket,
    SemanticSegmentationPacket,
    SpatialBbMappingPacket,
    TwoStagePacket,
    NNDataPacket
)
from depthai_sdk.oak_outputs.syncing import SequenceNumSync
from depthai_sdk.oak_outputs.xout.xout_base import XoutBase, StreamXout
from depthai_sdk.oak_outputs.xout.xout_depth import XoutDisparityDepth
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames
from depthai_sdk.oak_outputs.xout.xout_seq_sync import XoutSeqSync
from depthai_sdk.types import XoutNNOutputPacket
from depthai_sdk.visualize.bbox import BoundingBox
from depthai_sdk.visualize.colors import generate_colors, hex_to_bgr


class XoutNnData(XoutBase):
    def __init__(self, xout: StreamXout):
        self.nndata_out = xout
        super().__init__()

    def xstreams(self) -> List[StreamXout]:
        return [self.nndata_out]

    def new_msg(self, name: str, msg: dai.NNData) -> NNDataPacket:
        if name not in self._streams:
            return
        return NNDataPacket(name=self.get_packet_name(), nn_data=msg)


class XoutNnResults(XoutSeqSync, XoutFrames):
    def xstreams(self) -> List[StreamXout]:
        return [self.nn_results, self.frames]

    def __init__(self,
                 det_nn: 'NNComponent',
                 frames: StreamXout,
                 nn_results: StreamXout,
                 bbox: BoundingBox):
        self.det_nn = det_nn
        self.nn_results = nn_results

        XoutFrames.__init__(self, frames)
        XoutSeqSync.__init__(self, [frames, nn_results])

        self.name = 'NN results'
        self.labels = None
        self.bbox = bbox

        # TODO: add support for colors, generate new colors for each label that doesn't have colors
        if det_nn._labels:
            self.labels = []
            n_colors = [isinstance(label, str) for label in det_nn._labels].count(True)
            # List of colors in BGR format
            colors = generate_colors(n_colors)

            for label in det_nn._labels:
                if isinstance(label, str):
                    text = label
                    color = colors.pop(0)  # Take last row
                elif isinstance(label, list):
                    text = label[0]
                    color = hex_to_bgr(label[1])
                else:
                    raise ValueError('Model JSON config error. Label map list can have either str or list!')

                self.labels.append((text, color))

        self._resize_mode: ResizeMode = det_nn._ar_resize_mode
        self._nn_size: Tuple[int, int] = det_nn._size

    def package(self, msgs: Dict) -> XoutNNOutputPacket:
        nn_result = msgs[self.nn_results.name]
        img = msgs[self.frames.name]
        if type(nn_result) == dai.NNData:
            decode_fn = self.det_nn._decode_fn

            if decode_fn is None:
                return NnOutputPacket(self.get_packet_name(), img, nn_result, self.bbox)

            decoded_nn_result = decode_fn(nn_result)
            if type(decoded_nn_result) == Detections:
                packet = DetectionPacket(self.get_packet_name(), img, nn_result, self.bbox)
                return self._add_detections_to_packet(packet, decoded_nn_result)
            elif type(decoded_nn_result) == ImgLandmarks:
                return ImgLandmarksPacket(self.get_packet_name(), img, nn_result, decoded_nn_result, self.bbox)
            elif type(decoded_nn_result) == SemanticSegmentation:
                return SemanticSegmentationPacket(self.get_packet_name(), img, nn_result, decoded_nn_result, self.bbox)
            raise ValueError(f'NN result decoding failed! decode() returned type {type(nn_result)}')

        elif type(nn_result) in [dai.ImgDetections, dai.SpatialImgDetections]:
            packet = DetectionPacket(self.get_packet_name(), img, nn_result, self.bbox)
            return self._add_detections_to_packet(packet, nn_result)
        else:
            raise ValueError(f'Unknown NN result type: {type(nn_result)}')

    def _add_detections_to_packet(self,
                                  packet: DetectionPacket,
                                  dets: Union[dai.ImgDetections, dai.SpatialImgDetections, Detections]
                                  ) -> DetectionPacket:
        for detection in dets.detections:
            packet.detections.append(Detection(
                img_detection=detection if isinstance(detection, dai.ImgDetection) else None,
                label_str=self.labels[detection.label][0] if self.labels else str(detection.label),
                confidence=detection.confidence,
                color=self.labels[detection.label][1] if self.labels else (255, 255, 255),
                bbox=BoundingBox(detection),
                angle=detection.angle if hasattr(detection, 'angle') else None,
                ts=dets.getTimestamp()
            ))
        return packet


class XoutSpatialBbMappings(XoutDisparityDepth, SequenceNumSync):
    def __init__(self,
                 device: dai.Device,
                 stereo: dai.node.StereoDepth,
                 frames: StreamXout,  # passthroughDepth
                 configs: StreamXout,  # out
                 dispScaleFactor: float,
                 bbox: BoundingBox):
        self._stereo = stereo
        self.frames = frames
        self.configs = configs
        self.bbox = bbox

        XoutDisparityDepth.__init__(self, device, frames, dispScaleFactor, None)
        SequenceNumSync.__init__(self, 2)

    def new_msg(self, name: str, msg):
        # Ignore frames that we aren't listening for
        if name not in self._streams: return

        synced = self.sync(msg.getSequenceNum(), name, msg)
        if synced:
            return self.package(synced)

    def on_callback(self, packet) -> None:
        pass

    def xstreams(self) -> List[StreamXout]:
        return [self.frames, self.configs]

    def package(self, msgs: Dict) -> SpatialBbMappingPacket:
        return SpatialBbMappingPacket(
            self.get_packet_name(),
            msgs[self.frames.name],
            msgs[self.configs.name],
            disp_scale_factor=self.disp_scale_factor,
        )


class XoutTwoStage(XoutNnResults):
    """
    Two stage syncing based on sequence number. Each frame produces ImgDetections msg that contains X detections.
    Each detection (if not on blacklist) will crop the original frame and forward it to the second (stage) NN for
    inferencing.

    msgs = {
        '1': TwoStageSyncPacket(),
        '2': TwoStageSyncPacket(),
    }
    """

    def __init__(self,
                 det_nn: 'NNComponent',
                 second_nn: 'NNComponent',
                 frames: StreamXout,
                 det_out: StreamXout,
                 second_nn_out: StreamXout,
                 device: dai.Device,
                 input_queue_name: str,
                 bbox: BoundingBox):
        self.second_nn_out = second_nn_out
        # Save StreamXout before initializing super()!
        super().__init__(det_nn, frames, det_out, bbox)

        self.msgs: Dict[str, Dict[str, Any]] = dict()
        self.det_nn = det_nn
        self.second_nn = second_nn
        self.name = 'Two-stage detection'

        self.whitelist_labels: Optional[List[int]] = second_nn._multi_stage_nn.whitelist_labels
        self.scale_bb: Optional[Tuple[int, int]] = second_nn._multi_stage_nn.scale_bb

        self.device = device
        self.input_queue_name = input_queue_name
        self.input_queue = None
        self.input_cfg_queue = None

        self.lock = threading.Lock()

    def xstreams(self) -> List[StreamXout]:
        return [self.frames, self.nn_results, self.second_nn_out]

    # No need for `def visualize()` as `XoutNnResults.visualize()` does what we want

    def new_msg(self, name: str, msg: dai.Buffer):
        if name not in self._streams:
            return  # From Replay modules. TODO: better handling?

        # TODO: what if msg doesn't have sequence num?
        seq = str(msg.getSequenceNum())

        if seq not in self.msgs:
            self.msgs[seq] = dict()
            self.msgs[seq][self.second_nn_out.name] = []
            self.msgs[seq][self.nn_results.name] = None

        if name == self.second_nn_out.name:
            fn = self.second_nn._decode_fn
            if fn is not None:
                self.msgs[seq][name].append(fn(msg))
            else:
                self.msgs[seq][name].append(msg)

        elif name == self.nn_results.name:
            fn = self.det_nn._decode_fn
            if fn is not None:
                msg = fn(msg)

            self.add_detections(seq, msg)

            if self.input_queue_name:
                # We cannot create them in __init__ as device is not initialized yet
                if self.input_queue is None:
                    self.input_queue = self.device.getInputQueue(self.input_queue_name,
                                                                 maxSize=8,
                                                                 blocking=False)
                    self.input_cfg_queue = self.device.getInputQueue(self.input_queue_name + '_cfg',
                                                                     maxSize=8,
                                                                     blocking=False)

                for i, det in enumerate(msg.detections):
                    cfg = dai.ImageManipConfig()

                    if isinstance(det, dai.ImgDetection):
                        rect = (det.xmin, det.ymin, det.xmax, det.ymax)
                    else:
                        rect = det[0], det[1], det[2], det[3]

                    try:
                        angle = msg.angles[i]
                        if angle != 0.0:
                            rr = dai.RotatedRect()
                            rr.center.x = rect[0]
                            rr.center.y = rect[1]
                            rr.size.width = rect[2] - rect[0]
                            rr.size.height = rect[3] - rect[1]
                            rr.angle = angle
                            cfg.setCropRotatedRect(rr, normalizedCoords=True)
                        else:
                            cfg.setCropRect(rect)
                    except AttributeError:
                        cfg.setCropRect(rect)

                    cfg.setResize(*self.second_nn._size)

                    if i == 0:
                        try:
                            frame = self.msgs[seq][self.frames.name]
                        except KeyError:
                            continue

                        self.input_queue.send(frame)
                        cfg.setReusePreviousImage(True)

                    self.input_cfg_queue.send(cfg)

        elif name in self.frames.name:
            self.msgs[seq][name] = msg
        else:
            raise ValueError('Message from unknown stream name received by TwoStageSeqSync!')

        if self.synced(seq):
            # Frames synced!
            dets = self.msgs[seq][self.nn_results.name]
            packet = TwoStagePacket(
                self.get_packet_name(),
                self.msgs[seq][self.frames.name],
                dets,
                self.msgs[seq][self.second_nn_out.name],
                self.whitelist_labels,
                self.bbox
            )

            with self.lock:
                new_msgs = {}
                for name, msg in self.msgs.items():
                    if int(name) > int(seq):
                        new_msgs[name] = msg
                self.msgs = new_msgs

            return self._add_detections_to_packet(packet, dets)

    def add_detections(self, seq: str, dets: dai.ImgDetections):
        # Used to match the scaled bounding boxes by the 2-stage NN script node
        self.msgs[seq][self.nn_results.name] = dets

        if isinstance(dets, dai.ImgDetections):
            if self.scale_bb is None:
                return  # No scaling required, ignore

            for det in dets.detections:
                # Skip resizing BBs if we have whitelist and the detection label is not on it
                if self.labels and det.label not in self.labels: continue
                det.xmin -= self.scale_bb[0] / 100
                det.ymin -= self.scale_bb[1] / 100
                det.xmax += self.scale_bb[0] / 100
                det.ymax += self.scale_bb[1] / 100

    def synced(self, seq: str) -> bool:
        """
        Messages are in sync if:
            - dets is not None
            - We have at least one ImgFrame
            - number of recognition msgs is sufficient
        """
        packet = self.msgs[seq]

        if self.frames.name not in packet:
            return False  # We don't have required ImgFrames

        if not packet[self.nn_results.name]:
            return False  # We don't have dai.ImgDetections

        if len(packet[self.second_nn_out.name]) < self.required_recognitions(seq):
            return False  # We don't have enough 2nd stage NN results

        # print('Synced!')
        return True

    def required_recognitions(self, seq: str) -> int:
        """
        Required recognition results for this packet, which depends on number of detections (and white-list labels)
        """
        dets: List[dai.ImgDetection] = self.msgs[seq][self.nn_results.name].detections
        if self.whitelist_labels:
            return len([det for det in dets if det.label in self.whitelist_labels])
        else:
            return len(dets)



from enum import IntEnum
from typing import Union


class ResizeMode(IntEnum):
    """
    If NN input frame is in different aspect ratio than what the model expect, we have 3 different
    modes of operation. Full documentation:
    https://docs.luxonis.com/projects/api/en/latest/tutorials/maximize_fov/
    """
    LETTERBOX = 0  # Preserves full FOV by padding/letterboxing, but smaller frame means less features which might decrease NN accuracy
    STRETCH = 1  # Preserves full FOV, but frames are stretched to match the FOV, which might decrease NN accuracy
    CROP = 2  # Crops some FOV to match the required FOV, then scale. No potential NN accuracy decrease.
    FULL_CROP = 3  # No scaling is done, cropping is applied and FOV can be reduced by a lot

    # Parse string to ResizeMode
    @staticmethod
    def parse(mode: Union[str, 'ResizeMode']) -> 'ResizeMode':
        if isinstance(mode, ResizeMode):
            return mode

        mode = mode.lower()
        if mode == "letterbox":
            return ResizeMode.LETTERBOX
        elif mode == "stretch":
            return ResizeMode.STRETCH
        elif mode == "crop":
            return ResizeMode.CROP
        elif mode == "full_crop":
            return ResizeMode.FULL_CROP
        else:
            raise ValueError(f"Unknown resize mode {mode}! 'Options (case insensitive):" \
                             "STRETCH, CROP, LETTERBOX. Using default LETTERBOX mode.")



import depthai as dai
import numpy as np


def create_xyz(device: dai.Device, width: int, height: int):
    calibData = device.readCalibration()
    M_right = calibData.getCameraIntrinsics(dai.CameraBoardSocket.RIGHT, dai.Size2f(width, height))
    camera_matrix = np.array(M_right).reshape(3, 3)

    xs = np.linspace(0, width - 1, width, dtype=np.float32)
    ys = np.linspace(0, height - 1, height, dtype=np.float32)

    # generate grid by stacking coordinates
    base_grid = np.stack(np.meshgrid(xs, ys))  # WxHx2
    points_2d = base_grid.transpose(1, 2, 0)  # 1xHxWx2

    # unpack coordinates
    u_coord: np.array = points_2d[..., 0]
    v_coord: np.array = points_2d[..., 1]

    # unpack intrinsics
    fx: np.array = camera_matrix[0, 0]
    fy: np.array = camera_matrix[1, 1]
    cx: np.array = camera_matrix[0, 2]
    cy: np.array = camera_matrix[1, 2]

    # projective
    x_coord: np.array = (u_coord - cx) / fx
    y_coord: np.array = (v_coord - cy) / fy

    xyz = np.stack([x_coord, y_coord], axis=-1)
    return np.pad(xyz, ((0, 0), (0, 0), (0, 1)), "constant", constant_values=1.0)



import time
from datetime import timedelta, datetime
from pathlib import Path
from threading import Thread
from typing import Union, Callable, Dict, List

import depthai as dai

from depthai_sdk.classes import FramePacket
from depthai_sdk.components import Component
from depthai_sdk.logger import LOGGER
from depthai_sdk.recorders.video_recorder import VideoRecorder
from depthai_sdk.trigger_action.actions.abstract_action import Action

__all__ = ['RecordAction']


class RecordAction(Action):
    """
    Action that records video from specified inputs for a specified duration before and after trigger event.
    """

    def __init__(self,
                 inputs: Union[Component, Callable, List[Union[Component, Callable]]],
                 dir_path: str,
                 duration_before_trigger: Union[int, timedelta],
                 duration_after_trigger: Union[timedelta, int],
                 on_finish_callback: Callable[[Union[Path, str]], None] = None,
                 ):
        """
        Args:
            inputs: Inputs to record video from. Can be a single component or a list of components.
            dir_path: Path to directory where video files will be saved. Directory will be created if it doesn't exist.
            duration_before_trigger: Duration of video to record before trigger event.
            duration_after_trigger: Duration of video to record after trigger event.
            on_finish_callback: Callback function that will be called when recording is finished. Should accept a single argument - path to the folder with recorded video files.
        """
        super().__init__(inputs)
        self.path = Path(dir_path).resolve()
        if isinstance(duration_before_trigger, timedelta):
            duration_before_trigger = duration_before_trigger.total_seconds()
        if isinstance(duration_after_trigger, timedelta):
            duration_after_trigger = duration_after_trigger.total_seconds()
        if duration_before_trigger > 0 and duration_after_trigger > 0:
            self.duration_bt = duration_before_trigger
            self.duration_at = duration_after_trigger
        else:
            raise ValueError("Recording durations before and after trigger must be positive integers "
                             "or timedelta objects representing positive time difference")
        self.recorder = VideoRecorder()
        self.stream_names = []  # will be assigned during recorder setup
        self.buffers_status: Dict[str, Dict[str, Union[int, List[int]]]] = {}
        self.on_finish_callback = on_finish_callback

    def _run(self):
        """
        Starts recording video from specified inputs for a specified duration before and after trigger event.
        """
        while True:
            rbt_ids = self.buffers_status['ready']['before_t']

            # Wait until there is at least one ready before trigger buffer
            while not len(rbt_ids):
                time.sleep(0.01)
            rbt_id = rbt_ids.pop(0)

            # Create file for every stream
            buf_name = f'before_t_{rbt_id}'
            subfolder = f'{datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}'
            self.recorder.create_files_for_buffer(subfolder=subfolder, buf_name=buf_name)

            # Write before trigger buffer into file for every stream
            for stream_name in self.stream_names:
                while not self.recorder.is_buffer_empty(stream_name, buf_name):
                    self.recorder.write_from_buffer(wr_name=stream_name,
                                                    buf_name=buf_name,
                                                    n_elems=1)
                    time.sleep(0.01)

            # Wait until after trigger buffer with the same id is ready
            while rbt_id not in self.buffers_status['ready']['after_t']:
                time.sleep(0.01)

            # Write after trigger buffer into the same file for every stream
            buf_name = f'after_t_{rbt_id}'
            for stream_name in self.stream_names:
                while not self.recorder.is_buffer_empty(stream_name, buf_name):
                    self.recorder.write_from_buffer(wr_name=stream_name,
                                                    buf_name=buf_name,
                                                    n_elems=1)
                    time.sleep(0.01)

            # Close files
            LOGGER.debug(f'Saved to {str(self.path / subfolder)}')
            self.recorder.close_files()

            # Call on_finish_callback if it is specified
            if self.on_finish_callback is not None:
                self.on_finish_callback(str(self.path / subfolder))

    def activate(self):
        # Setup for the current recording
        wbt_id = self.buffers_status['writing']['before_t']
        self.buffers_status['ready']['before_t'].append(wbt_id)
        self.buffers_status['writing']['after_t'].append(wbt_id)

        # Setup for the next recording
        new_id = wbt_id + 1
        buffers = {f'before_t_{new_id}': self.duration_bt,
                   f'after_t_{new_id}': self.duration_at}

        self.recorder.init_buffers(buffers)
        self.buffers_status['writing']['before_t'] = new_id

    def on_new_packets(self, packets: Dict[str, FramePacket]):
        # Write into the only writing before trigger buffer
        wbt_id = self.buffers_status['writing']['before_t']

        # Extract imgFrames from packets --> Syncing packets is redundant for RecordAction, just sync frames
        frames = dict.fromkeys(self.stream_names)
        for name, _ in frames.items():
            for packet_name, packet in packets.items():
                if name in packet_name:
                    frames[name] = packet.msg
                    packets.pop(packet_name)
                    break

        for _, frame in frames.items():
            if frame is None:
                raise Exception("Extracting msg from packets failed, check if all streams are present")

        self.recorder.add_to_buffers(f'before_t_{wbt_id}', frames)  # does for every stream

        # Write into all writing after trigger buffers
        wat_ids = self.buffers_status['writing']['after_t']
        for i, buffer_id in enumerate(list(wat_ids)):
            buf_name = f'after_t_{buffer_id}'
            all_streams_ready = True
            for stream_name in self.stream_names:
                if not self.recorder.is_buffer_full(stream_name, buf_name):
                    self.recorder.add_to_buffer(stream_name, buf_name, frames[stream_name])
                    all_streams_ready = False
            if all_streams_ready:
                wat_ids.pop(i)
                self.buffers_status['ready']['after_t'].append(buffer_id)

    def setup(self, device: dai.Device, xouts: List['XoutFrames']):
        self.stream_names = [xout.name for xout in xouts]  # e.g., [color_video, color_bitstream]
        LOGGER.debug(f'RecordAction: stream_names = {self.stream_names}')
        self.recorder.update(self.path, device, xouts)
        self._run_thread()

    def _run_thread(self):
        buffers = {'before_t_1': self.duration_bt, 'after_t_1': self.duration_at}
        self.recorder.init_buffers(buffers)  # does for every stream
        self.buffers_status = {'writing': {'before_t': 1, 'after_t': []},
                               # there is always only one writing before trigger buffer
                               'ready': {'before_t': [], 'after_t': []}}

        process = Thread(target=self._run)
        process.start()
